"""Exceptions for the kegscraping package"""

class UnclosedJSONError(Exception):
    """
    Raised when a JSON string is never closed.
    """
    pass

class NotFound(Exception):
    pass

class ParseFailure(Warning):
    pass

class Unauthorised(Exception):
    pass

class ServerError(Exception):
    pass

class TimeOut(Exception):
    pass

class UnimplementedWarning(Warning):
    pass

class WebServiceError(Exception):
    pass


"""
Module for reading the event list pdf at https://www.kegs.org.uk/eventsPDF.cfm
"""
from typing import Final
from io import BytesIO
import re

import dateparser
from pypdf import PdfReader

from .events import ATCEvent

MONTHS: Final[tuple[str, str, str, str, str, str, str, str, str, str, str, str]] = (
    "January", "February", "March", "April", "May", "June", "July",
    "August", "September", "October", "November", "December")

# noinspection PyTypeChecker
re_mnth = '|'.join(map(str.upper, MONTHS))


def read(pdf_data: bytes) -> list[ATCEvent]:
    """
    Parse a KEGS Events pdf into a list of events. Does not validate the pdf.
    :param pdf_data: The pdf as bytes
    :return: a list of events
    """
    pdf = PdfReader(BytesIO(pdf_data))
    text = ''
    for page in pdf.pages:
        text += page.extract_text() + '\n'

    events: list[ATCEvent] = []
    year = None
    for line in text.split('\n'):
        line: str
        line = line.strip()

        if line == "KEGS Events":
            # It's the top line
            ...

        elif m := re.match(fr"({re_mnth}) \d\d\d\d", line):
            # In format: {MONTH} {yyyy}
            m = m.group()

            # We can ignore the month, but we can't get the year data from anywhere else
            year = m[-4:]

        elif line == '':
            # Ignore blank lines
            continue

        else:
            # We can assume that the rest are just events

            # Format: [Day name in titlecase], [day #] [1st 3 chars of month] [All Day | hh:mm - hh:mm] [Description]
            split = line.split(', ')
            day_name, rest = split[0], ', '.join(split[1:])

            split = rest.split(' ')
            day, month, rest = split[0], split[1], ' '.join(split[2:])

            for full_month in MONTHS:
                if full_month.startswith(month):
                    month = full_month
            month = MONTHS.index(month) + 1

            all_day = rest.startswith("All Day ")
            if all_day:
                title = rest[len("All Day "):]
                start, end = '', ''
            else:
                # Some time in format: hh:mm
                split = rest.split(' ')

                start, end, title = split[0], split[2], ' '.join(split[3:])

            start_date = dateparser.parse(f"{day}/{month}/{year} {start}")
            end_date = dateparser.parse(f"{day}/{month}/{year} {end}")

            events.append(ATCEvent(start_date, end_date, title=title, all_day=all_day))

    return events



"""Post, Discussion and Forum classes"""
from __future__ import annotations

from urllib.parse import urlparse, parse_qs
from datetime import datetime

import dateparser
from dataclasses import dataclass, field
from bs4 import BeautifulSoup, NavigableString, PageElement

from . import session, user


@dataclass
class Post:
    """Represents a post in a discussion in a forum on the kegsnet website"""

    id: int = None
    """Actually the integer in the url fragment that links to this post, which is attached to the discussion link"""
    creator: user.User = None
    date: datetime = None
    title: str = None
    content: str = field(repr=False, default=None)

    _discussion: Discussion = None
    _session: session.Session = field(repr=False, default=None)

    def update_from_html(self, elem: PageElement):
        # --- Data from the <header> tag
        header = elem.find("header")
        hdata = header.find("div", {"class": "flex-column"})

        self.title = hdata.find("h3").text

        user_anchor = hdata.find("a")
        user_url = user_anchor.attrs.get("href")
        uid = int(
            parse_qs(
                urlparse(user_url).query
            ).get("id")[0]
        )

        self.creator = self._session.connect_user_by_id(uid)
        # We can actually provide the name of the creator, even if other data is inaccessible
        self.creator.name = user_anchor.text

        self.date = dateparser.parse(
            header.find("time").text
        )

        # Other data
        permalink = elem.find("a", {"title": "Permanent link to this post"}).attrs.get("href")
        parse = urlparse(permalink)
        self.id = int(parse.fragment[1:])

        self.content = str(elem.find("div", {"class": "post-content-container"}))


@dataclass
class Discussion:
    """Represents a discussion within a forum on the kegsnet website"""
    id: int = None

    name: str = None
    # author: user.User = None # It only shows a name & pfp - but not an actual link

    date_created: datetime = None
    # last_post: Post = None
    reply_count: int = None

    _forum: Forum = field(repr=False, default=None)
    _session: session.Session = field(repr=False, default=None)
    _top_post: Post = field(repr=False, default=None)

    posts: list[Post] = field(repr=False, default_factory=lambda: [])

    def update_from_forum_html(self, elem: PageElement) -> None:
        """
        Update the discussion from HTML on a forum's page
        :param elem: the HTML data as a bs4.PageElement object
        """
        for i, item in enumerate(elem.find_all("td")):
            if i == 0:
                # Star this discussion
                ...

            elif i == 1:
                # Name
                anchor = item.find("a")
                self.name = anchor.text

                # You can also get id from the url
                parse = urlparse(anchor.attrs.get("href"))
                qparse = parse_qs(parse.query)
                self.id = int(qparse.get("d")[0])

            elif i == 2:
                # Started by
                ...

            elif i == 3:
                # Reply count
                self.reply_count = int(item.find("a").text.strip())

            elif i == 4:
                # Last post
                ...

            elif i == 5:
                # Date created
                text = item.text.strip()
                self.date_created = dateparser.parse(text)

            else:
                break

    def update(self):
        """
        Update the discussion from the corresponding url. Requires an id
        :return:
        """
        response = self._session.rq.get("https://vle.kegs.org.uk/mod/forum/discuss.php",
                                        params={
                                    "d": self.id,
                                    "mode": 1
                                })

        soup = BeautifulSoup(response.text, "html.parser")

        self.name = soup.find("h3", {"class": "discussionname"}).text
        core_attrs = {"data-region-content": "forum-post-core"}

        post_htmls = soup.find_all("div", core_attrs)
        top_post_html = soup.find("div", {"class": "firstpost"}).find("div", core_attrs)

        self._top_post = Post(_discussion=self, _session=self._session)
        self._top_post.update_from_html(top_post_html)

        self.posts = []
        for post_html in post_htmls:
            self.posts.append(Post(_discussion=self, _session=self._session))
            self.posts[-1].update_from_html(post_html)

    @property
    def url(self) -> str:
        """Get the url of this discussion"""
        return f"https://vle.kegs.org.uk/mod/forum/discuss.php?d={self.id}"

    @property
    def top_post(self) -> Post:
        """
        Fetch the first post in a discussion
        """
        if not self._top_post:
            self.update()

        return self._top_post


@dataclass
class Forum:
    """Represents a forum on KEGSNET - e.g. the news forum"""
    id: int

    name: str = None
    description: str = None
    contents: list[Discussion] = None

    _session: session.Session = field(repr=False, default=None)

    def update_by_id(self):
        """Update attributes by requesting the corresponding webpage. Requires an id"""
        response = self._session.rq.get("https://vle.kegs.org.uk/mod/forum/view.php",
                                        params={"f": self.id})
        soup = BeautifulSoup(response.text, "html.parser")

        container = soup.find("div", {"role": "main"})
        for i, element in enumerate(container.children):
            if element.name == "h2":
                self.name = element.text

            elif element.name == "div":
                div_id = element.attrs.get("id")

                if div_id == "intro":
                    self.description = element.text

                else:
                    element: PageElement
                    post_list = element.find("table", {"class": "table table-hover table-striped discussion-list"})

                    for tpart in post_list.children:
                        tpart: PageElement
                        if tpart.name == "tbody":
                            discussions = []

                            # List of discussions
                            for discuss_elem in tpart.children:
                                if not isinstance(discuss_elem, NavigableString):
                                    discussion = Discussion(_forum=self, _session=self._session)
                                    discussion.update_from_forum_html(discuss_elem)
                                    discussions.append(discussion)

                            self.contents = discussions



from __future__ import annotations

from dataclasses import dataclass

from . import session

@dataclass
class CourseCategory:
    name: str = None # e.g. 'Chemistry Y9-11'

    _sess: session.Session = None


from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

from . import session

@dataclass
class Organisation:
    pages_graph: list[int] = None

    trees: float = None
    co2: int = None
    energy: float = None
    since: datetime = None

    sess: session.Session = None



from __future__ import annotations

import dateparser
from bs4 import BeautifulSoup
from typing import Final
from dataclasses import dataclass, field
import warnings
from datetime import datetime

from . import session

DELETED_USER: Final[str] = "This user account has been deleted"
INVALID_USER: Final[str] = "Invalid user"
FORBIDDEN_USER: Final[str] = "The details of this user are not available to you"


@dataclass
class User:
    id: int = None

    name: str = None
    email: str = field(repr=False, default=None)
    image_url: str = field(repr=False, default=None)

    country: str = field(repr=False, default=None)
    city: str = field(repr=False, default=None)
    web_page: str = field(repr=False, default=None)

    interests: list = field(repr=False, default=None)
    courses: list = field(repr=False, default=None)

    first_access: datetime = field(repr=False, default=None)
    last_access: datetime = field(repr=False, default=None)

    description: str = field(repr=False, default=None)

    _session: session.Session = field(repr=False, default=None)

    flags: list[str] = field(repr=False, default_factory=list)

    @property
    def has_default_image(self) -> bool | None:
        if self.image_url is None:
            return None

        return self.image_url == "https://vle.kegs.org.uk/theme/image.php/trema/core/1585328846/u/f1"

    @property
    def profile_image(self) -> bytes:
        return self._session.rq.get(self.image_url).content

    def update_from_id(self):
        response = self._session.rq.get("https://vle.kegs.org.uk/user/profile.php",
                                        params={"id": self.id})
        text = response.text
        soup = BeautifulSoup(text, "html.parser")

        self.flags = []

        if DELETED_USER in text:
            self.flags.append(DELETED_USER)
            warnings.warn(f"User id {self.id} is deleted!")

        elif INVALID_USER in text:
            self.flags.append(INVALID_USER)
            warnings.warn(f"User id {self.id} is invalid!")

        elif FORBIDDEN_USER in text:
            self.flags.append(FORBIDDEN_USER)
            warnings.warn(f"User id {self.id} is forbidden!")

        else:
            # Get user's name
            self.name = str(soup.find("div", {"class": "page-header-headings"}).contents[0].text)

            # Get user image
            self.image_url = soup.find_all("img", {"class": "userpicture"})[1].get("src")

            user_profile = soup.find("div", {"class": "userprofile"})
            self.description = user_profile.find("div", {"class": "description"})

            categories = user_profile.find_all("section", {"class", "node_category"})

            interests_node, interests, courses = None, [], []

            for category in categories:
                category_name = category.find("h3").contents[0]

                if category_name == "User details":
                    user_details = list(category.children)[1]

                    # This is an unordered list containing the Email, Country, City and Interest
                    content_nodes = user_details.find_all("li", {"class", "contentnode"})

                    for li in content_nodes:
                        dl = li.find("dl")

                        dd = dl.find("dd")
                        item_name = dl.find("dt").contents[0]

                        if item_name == "Email address":
                            self.email = dl.find("a").contents[0]

                        elif item_name == "City/town":
                            self.city = dd.contents[0]

                        elif item_name == "Country":
                            self.country = dd.contents[0]

                        elif item_name == "Web page":
                            self.web_page = dl.find("a").get("href")

                        elif item_name == "Interests":
                            interests_node = dl

                    if interests_node is not None:
                        try:
                            for anchor in interests_node.find_all("a"):
                                interests.append(anchor.contents[0][21:])
                        except IndexError:
                            ...

                        if interests:
                            self.interests = interests

                elif category_name == "Course details":
                    for anchor in category.find_all("a"):
                        courses.append((anchor.get("href").split('=')[-1],
                                        anchor.contents[0]))
                    if courses:
                        self.courses = courses

                elif category_name == "Miscellaneous":
                    ...

                elif category_name == "Reports":
                    ...

                elif category_name == "Login activity":
                    for i, activity in enumerate(category.find_all("dd")):
                        date_str = activity.contents[0]
                        date_str = date_str[:date_str.find('(')]

                        if i == 0:
                            self.first_access = dateparser.parse(date_str)
                        else:
                            self.last_access = dateparser.parse(date_str)



from __future__ import annotations

import atexit
import warnings
from dataclasses import dataclass, field

import requests
from bs4 import BeautifulSoup

from . import course

from ..util import commons


@dataclass
class Session:
    id: str = None
    email: str = None

    username: str = field(repr=False, default=None)
    first_name: str = field(repr=False, default=None)
    last_name: str = field(repr=False, default=None)
    display_name: str = field(repr=False, default=None)
    institution_code: str = field(repr=False, default=None)
    account_type: str = field(repr=False, default=None)

    rq: requests.Session | None = field(repr=False, default=None)

    def __post_init__(self):
        atexit.register(self.logout)

    def connect_courses(self):
        data = self.rq.get("https://www.kerboodle.com/api/v2/courses").json()

        ret = []
        data = data.get("data", [])
        for course_data in data:
            match course_data["type"]:
                case "course":
                    attrs = course_data["attributes"]
                    ret.append(course.Course(_sess=self, **attrs))

                case _:
                    warnings.warn(f"Unknown course type: {course_data}. "
                                  f"Please file an issue on github: https://github.com/BigPotatoPizzaHey/kegscraper/issues")

        return ret

    def connect_course_by_id(self, _id: int):
        return course.Course(id=_id, _sess=self)

    def update_by_settings_api(self):
        # There is a huge amount of information in this response.
        # Only some of it is useful.
        data = self.rq.get("https://www.kerboodle.com/api/v2/settings").json()["data"]

        attrs = data["attributes"]
        # possible useful attrs of attrs:
        # - client
        # - profile["school"]

        # --- #
        profile = attrs["profile"]

        self.id = profile["userid"] # profile["id"] is something weird
        self.account_type = profile["type"]
        self.first_name = profile["first_name"]
        self.last_name = profile["last_name"]
        self.display_name = profile["display_name"]
        self.email = profile["email"]
        self.username = profile["username"]

        self.institution_code = profile["school"]["code"]

    def logout(self):
        """Send a logout request to kerboodle. Might not have any effect"""
        resp = self.rq.get("https://www.kerboodle.com/app")
        soup = BeautifulSoup(resp.text, "html.parser")

        resp = self.rq.post("https://www.kerboodle.com/users/logout", data={
            "_method": "delete",
            "authenticity_token": soup.find("meta", {"name": "csrf-token"}).get("content")
        })

        print(f"Logged out with status code {resp.status_code}")

        self.rq = None

def login(institution_code: str, username: str, password: str, auto_update: bool = True) -> Session:
    rq = requests.Session()
    resp = rq.get("https://www.kerboodle.com/users/login")
    soup = BeautifulSoup(resp.text, "html.parser")

    qs = commons.eval_inputs(soup)

    del qs["commit"]
    qs.update({
        "user[login]": username,
        "user[password]": password,
        "user[institution_code]": institution_code
    })

    rq.post("https://www.kerboodle.com/users/login", data=qs)
    sess = Session(rq=rq, institution_code=institution_code)

    if auto_update:
        sess.update_by_settings_api()

    return sess



"""
Timetable Dataclasses for bromcom
"""
from __future__ import annotations

from datetime import datetime
from dataclasses import dataclass

from ..util import commons

from . import session

@dataclass
class WeekDate:
    """
    A dataclass representing a start of a week to be used with the bromcom timetable.
    """
    term_i: int
    week_i: int
    date: datetime
    _sess: session.Session = None

@dataclass
class Lesson:
    """
    A dataclass representing a lesson in the bromcom timetable
    """
    period: str
    subject: str
    class_name: str
    room: str
    teacher: str
    teacher_id: int
    # ^^ I am unaware of any use. Perhaps if they have the same name
    week_id: int
    # ^^ Seems to be that the only use is to differentiate week a/b?

    start: datetime
    end: datetime

    color: str = None

    _sess: session.Session = None
    week_a_b: str = None

    @property
    def weekday(self):
        return self.start.strftime("%A")

    # @property
    # def week_a_b(self):
    #     return "ba"[self.week_id % 2]

def get_mode_timetable(_timetable: list[Lesson]) -> dict[str, dict[str, Lesson]]:
    def find_lessons(_period: str = None, day_name: str = None) -> list[Lesson]:
        ret = []
        for _lesson in _timetable:
            if _lesson.period == _period or _period is None:
                if _lesson.start.strftime("%A") == day_name or day_name is None:
                    ret.append(_lesson)
        return ret

    periods = []
    for lesson in _timetable:
        if lesson.period in periods:
            break
        periods.append(lesson.period)

    weekdays = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday']
    result = {}

    for weekday in weekdays:
        result[weekday] = {}
        for period in periods:
            lessons = find_lessons(period, weekday)
            result[weekday][period] = commons.get_mode(lessons, no_dunder=True)

    return result


from __future__ import annotations

import mimetypes

import dateparser
import requests
from bs4 import BeautifulSoup
from datetime import datetime

from dataclasses import dataclass

from ..util import commons

from . import org

@dataclass
class Session:
    username: str = None

    balance: float = None
    pages: int = None
    jobs: int = None

    trees: float = None
    co2: int = None  # Grams of co2
    energy: float = None  # joules
    since: datetime = None

    pages_graph: list[int] = None  # idx -1 is this week, -2 is last week, etc. (reverse?)

    organisation: org.Organisation = None

    rq: requests.Session = None

    def update_by_env(self):
        """
        Reassign session attributes (esp. self.organisation) by making a request to the environment dashboard
        """
        response = self.rq.get(f"http://printing.kegs.local:9191/environment/dashboard/{self.username}")

        self.update_by_env_dash_html(BeautifulSoup(response.text, "html.parser"))

    def logout(self):
        """
        Supposedly send a logout request. Probably doesn't do anything
        """
        self.rq.get("http://printing.kegs.local:9191/app?service=direct/1/UserSummary/$UserBorder.logoutLink")

    def update_from_dashboard(self):
        """
        Reassign session attributes by making a request to the main dashboard
        """
        resp = self.rq.get("http://printing.kegs.local:9191/app?service=page/UserSummary")
        self.update_by_dash_html(BeautifulSoup(resp.text, "html.parser"))

    def __post_init__(self):
        self.organisation = org.Organisation(sess=self)

    def update_by_env_dash_html(self, soup: BeautifulSoup):
        """
        Reassign session attributes (esp. self.organisation) by using the environment dashboard
        :param soup: Beautifulsoup representing the environment dashboard
        """
        # maybe scrape sheets (week/month), cost/month + trees/co2/energy

        # graph (canvas) - have to scrape js!
        for script in soup.find_all("script", {"type": "text/javascript"}):
            if script.contents:
                content = script.contents[0]
                find_str = "datasets : ["
                if find_str in content:
                    i = content.find(find_str)
                    content = content[i:]

                    find_str = "data : ["
                    f1 = content.find(find_str)
                    f2 = content[f1 + 8:].find(find_str) + f1 + 8

                    self.organisation.pages_graph = commons.consume_json(content, f1 + 6)  # num. pages
                    self.pages_graph = commons.consume_json(content, f2 + 6)  # num. pages

        for div in (soup.find("div", {"class": "box box50-100 medium"}),
                    soup.find("div", {"class": "box box50-100 darker"})):
            h2 = div.find("h2", {"class": "centered"})

            if h2.text.strip() == "Organization Impact":
                for stat in div.find_all("div", {"class": "env-stats-text"}) + \
                            div.find_all("div", {"class": "centered env-impact"}):
                    stat = stat.text.strip()

                    if stat.endswith(" trees"):
                        self.organisation.trees = float(stat.split()[0])
                    elif stat.endswith(" kg of CO2"):
                        self.organisation.co2 = 1000 * int(stat.replace(',', '').split()[0])
                    elif stat.endswith(" bulb hours"):
                        self.organisation.energy = 60 * 60 * 60 * int(stat.replace(',', '').split()[0])
                    elif stat.startswith("Since\n"):
                        self.organisation.since = dateparser.parse(stat[len("Since\n"):])

    def update_by_dash_html(self, soup: BeautifulSoup):
        """
        Reassign session attributes using the main dashboard html page
        :param soup: Beautifulsoup representing the html
        """
        self.username = soup.find("span", {"id": "username"}).text

        bal_div = soup.find("div", {"class": "widget stat-bal"})
        self.balance = float(bal_div.find("div", {"class": "val"}).text.strip()[1:])  # [1:] is to remove '£'

        pages_div = soup.find("div", {"class": "widget stat-pages"})
        self.pages = int(pages_div.find("div", {"class": "val"}).text)

        jobs_div = soup.find("div", {"class": "widget stat-jobs"})
        self.jobs = int(jobs_div.find("div", {"class": "val"}).text)

        env_div = soup.find("div", {"id": "enviro", "class": "col"})
        env_widget = env_div.find("div", {"class": "widget"})

        for li in env_widget.find_all("li"):
            key = li.get("class")[0]
            val = li.text.strip()

            match key:
                case "trees":
                    self.trees = float(val.split('%')[0]) / 100  # /100 because it's a percentage
                case "co2":
                    self.co2 = int(val.split('g')[0])
                case "energy":
                    val = float(val.split("hours")[0])  # hours running a 60W light bulb
                    val = val * 60 * 60  # seconds spent powering bulb
                    val *= 60  # 60W bulb
                    self.energy = val
                case "since-date":
                    val = val.replace("Since", '').strip()
                    self.since = dateparser.parse(val)

    def get_balance_graph(self, width: int=668, height: int=400) -> tuple[str, bytes]:
        resp = self.rq.get(f"http://printing.kegs.local:9191/app?service=chart/UserSummary/{width}/{height}/$Chart")
        # &sp=SBalance+history+for+{self.username}&69402papercut-mf")
        # it appears that the last part of the url is unnecessary

        return mimetypes.guess_extension(resp.headers["Content-Type"]), resp.content

def login(username: str, password: str) -> Session:
    """
    Login to papercut mf
    :param username:
    :param password:
    :return: A session object
    """
    sess = requests.Session()

    # Make an initial request (to set cookies)
    sess.get("http://printing.kegs.local:9191/user")

    # These headers were copied directly from my browser (no cookies)
    # Some of these headers have to be removed so that it works for other pages
    sess.headers = {
        # "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        # "Accept-Encoding": "gzip, deflate",
        "Accept-Language": "en-US,en;q=0.9",
        "Cache-Control": "max-age=0",
        "Connection": "keep-alive",
        # "Content-Length": "302",
        # "Content-Type": "application/x-www-form-urlencoded",
        "Host": "printing.kegs.local:9191",
        # "Origin": "http://printing.kegs.local:9191",
        "Referer": "http://printing.kegs.local:9191/app",
        "Upgrade-Insecure-Requests": "1",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36"
    }

    resp = sess.post("http://printing.kegs.local:9191/app",
                     data={
                         "service": "direct/1/Home/$Form",
                         "sp": "S0",
                         "Form0": "$Hidden$0,$Hidden$1,inputUsername,inputPassword,$Submit$0,$PropertySelection",

                         "$Hidden$0": "true",
                         "$Hidden$1": "X",
                         "inputUsername": username,
                         "inputPassword": password,

                         "$Submit$0": "Log in",
                         "$PropertySelection": "en"
                     })

    ret = Session(rq=sess, username=username)
    # Since we receive the html of the main dashboard as the response content, we might as well parse it
    ret.update_by_dash_html(BeautifulSoup(resp.text, "html.parser"))

    return ret



"""
Class representing tags accessible through here: https://vle.kegs.org.uk/tag/index.php?tag=woodlouse
"""
from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from urllib.parse import urlparse, parse_qs
from typing import Self

import dateparser
import requests
from bs4 import BeautifulSoup, PageElement
from . import session, user, blog
from ..util import commons


@dataclass
class Tag:
    name: str = None
    exists: bool = None

    description: PageElement = field(default=None, repr=False)
    related_tags: list[Tag] = field(default_factory=list)

    id: int = None
    flagged_inappropriate: bool = field(repr=False, default=False)

    item_id: str = None

    _session: session.Session = field(repr=False, default=None)

    @classmethod
    def from_json(cls, data: dict, _sess: session.Session) -> Self:
        return cls(
            flagged_inappropriate=bool(data["flag"]),
            id=data["id"],
            item_id=data["itemid"],
            name=data["name"], # == rawname?
            exists=True,
            # instance id, context id?
            _session=_sess
        )

    @property
    def url(self):
        if self.id:
            return f"https://vle.kegs.org.uk/tag/index.php?id={self.id}"
        elif self.name:
            return f"https://vle.kegs.org.uk/tag/index.php?tag={self.name}"
        else:
            warnings.warn(f"Could not infer a Tag id for {self}")
            return ""

    def _update_from_response(self, response: requests.Response):
        self.exists = response.url != "https://vle.kegs.org.uk/tag/search.php"
        if self.exists:
            soup = BeautifulSoup(response.text, "html.parser")
            main = soup.find("div", {"role": "main"})

            self.name = main.find("h2").text

            mng_box = main.find("div", {"class": "tag-management-box"})
            tedit = mng_box.find("a", {"class": "edittag"})
            href = tedit.attrs.get("href", '')
            q_parse = parse_qs(urlparse(href).query)
            self.id = int(q_parse.get("id")[0])

            # get desc
            self.description = main.find("div", {"class": "tag-description"})

            # get related tags
            self.related_tags.clear()
            related_tag_div = main.find("div", {"class": "tag-relatedtags"})
            if related_tag_div:
                for li in related_tag_div.find_all("li"):
                    href = li.find("a").attrs["href"]
                    if href == '#':
                        continue

                    parsed = urlparse(href)
                    q_parse = parse_qs(parsed.query)
                    related_tag_name = q_parse["tag"][0]
                    self.related_tags.append(
                        Tag(
                            related_tag_name,
                            _session=self._session
                        )
                    )

    def update(self):
        """
        Update by name or id
        """
        response = self._session.rq.get(self.url)
        self._update_from_response(response)

    def connect_interested_users(self, limit: int = 5, offset: int = 0):
        users = []

        for page in commons.generate_page_range(limit, offset, 5, 0)[0]:
            data = self._session.webservice("core_tag_get_tagindex", tagindex={
                "tc": 1,
                "tag": self.name,
                "ta": 1,  # 1 = users, 3 = courses, 7 = blog posts.
                "page": str(page)
            })

            soup = BeautifulSoup(data["content"], "html.parser")

            lis = soup.find_all("li", {"class": "media"})
            if not lis:
                break # if it is empty, don't need to make more webreqs

            for li in lis:
                a = li.find("a")
                href = a.attrs["href"]
                q_parse = parse_qs(urlparse(href).query)

                uid = int(q_parse["id"][0])

                img = a.find("img")
                src = img.attrs["src"]

                body = li.find("div", {"class": "media-body"})
                name = body.text.strip()

                users.append(user.User(id=uid, name=name, image_url=src, _session=self._session))

        return users

    def connect_tagged_blog_entries(self, limit: int = 5, offset: int = 0):
        entries = []

        for page in commons.generate_page_range(limit, offset, 5, 0)[0]:
            data = self._req_get_tagindex(page, 7).json()[0]["data"]["content"]
            soup = BeautifulSoup(data, "html.parser")

            for li in soup.find_all("li", {"class": "media"}):
                a = li.find("a")
                href = a.attrs["href"]
                q_parse = parse_qs(urlparse(href).query)

                uid = int(q_parse["id"][0])

                img = a.find("img")
                src = img.attrs["src"]

                body = li.find("div", {"class": "media-body"})

                entry_a = body.find("a")
                subject = entry_a.contents[0].strip()

                href = entry_a.attrs["href"]
                q_parse = parse_qs(urlparse(href).query)
                entry_id = int(q_parse["entryid"][0])

                muted = body.find("div", {"class": "muted"})
                split = muted.text.split(',')
                author_name = split[0].strip()
                date = dateparser.parse(','.join(split[1:]))

                author = user.User(id=uid, name=author_name, image_url=src)

                entries.append(
                    blog.Entry(
                        id=entry_id,
                        subject=subject,
                        date_created=date,
                        author=author,
                        _session=self._session
                    )
                )

        return entries

    def edit(self, new_description: str | PageElement = None, related_tags: list[Tag | str] = None):
        if new_description is None:
            new_description = self.description
        if related_tags is None:
            related_tags = self.related_tags

        if not isinstance(new_description, str):
            if new_description is not None:
                new_description = new_description.prettify()
            # else:
            #     new_description = ''

        related_tags = list(map(lambda x: x if isinstance(x, str) else x.name, related_tags))

        response = self._session.rq.get("https://vle.kegs.org.uk/tag/edit.php", params={"id": self.id})

        soup = BeautifulSoup(response.text, "html.parser")

        # it appears that order matters??? Not sure
        data = [
                   ("id", self.id),
                   ("returnurl", ''),
                   ("sesskey", self._session.sesskey),
                   ("_qf__tag_edit_form", 1),
                   ("mform_isexpanded_id_tag", new_description),
                   ("description_editor[text]", ''),
                   ("description_editor[format]", 1),
                   ("description_editor[itemid]", soup.find("input", {"name": "description_editor[itemid]"})),
                   ("relatedtags", "_qf__force_multiselect_submission"),
               ] + [("relatedtags[]", tagname) for tagname in related_tags] + [
                   ("submitbutton", "Update"),
               ]

        return self._session.rq.post("https://vle.kegs.org.uk/tag/edit.php",
                                     data=data)



"""
File class
"""
from __future__ import annotations

import os.path

from dataclasses import dataclass
from datetime import datetime
from typing import Self

from . import session
from . import user as _user


@dataclass
class File:
    """
    Class representing both files and directories in kegsnet
    """
    name: str = None
    path: str = None

    size: int = None
    author: str = None
    license: str = None

    mime: str = None
    type: str = None

    url: str = None
    icon_url: str = None

    datemodified: datetime = None
    datecreated: datetime = None

    user: _user.User = None
    is_external: bool = False

    _session: session.Session = None

    def __repr__(self):
        return f"<{self.type.title()}: {os.path.join(self.path, self.name)}>"

    @property
    def contents(self) -> list[File] | bytes:
        """
        Retrieve contents of the file or directory
        :return: list of files for directories, or file content as bytes
        """
        if self.is_dir:
            # Get the folder contents
            return self._session.files_in_dir(self.path)
        else:
            return self._session.rq.get(self.url).content

    def delete(self):
        """
        Deletes the file from the session's file manager
        """
        self._session.rq.post("https://vle.kegs.org.uk/repository/draftfiles_ajax.php",
                              params={"action": "delete"},
                              data={
                                  "sesskey": self._session.sesskey,

                                  "clientid": self._session.file_client_id,
                                  "itemid": self._session.file_item_id,
                                  "filename": self.name,
                                  "filepath": self.path
                              })
        self._session.file_save_changes()

    @classmethod
    def from_json(cls, data: dict, _session: session.Session = None) -> Self:
        """Load a file from JSON data"""
        return cls(name=data.get("filename"),
                   path=data.get("filepath"),
                   size=data.get("size"),
                   author=data.get("author"),
                   license=data.get("license"),
                   mime=data.get("mimetype"),
                   type=data.get("type"),
                   url=data.get("url"),
                   icon_url=data.get("icon"),
                   datemodified=datetime.fromtimestamp(data.get("datemodified")),
                   datecreated=datetime.fromtimestamp(data.get("datecreated")),
                   user=_session.connected_user,
                   _session=_session)

    @classmethod
    def from_json2(cls, data: dict, _sess: session.Session) -> Self:
        """Load a file from JSON data in a slightly different format"""
        return cls(
            name=data.get("filename"),
            path=data.get("filepath"),
            size=data.get("filesize"),
            url=data.get("fileurl"),
            is_external=data.get("isexternalfile"),
            mime=data.get("mimetype"),
            type="file",
            datemodified=datetime.fromtimestamp(data.get("timemodified")),
            _session=_sess
        )

    @property
    def is_dir(self):
        """Check if the file is actually a directory"""
        return self.type == "folder"



from __future__ import annotations

from ..util import commons


def download_header(idx: int = 1):
    """
    Get a https://it.kegs.org.uk background header image (jpg) by idx
    :param idx: int 1-8 inclusive
    :return: (bytes) jpg image
    """
    if not 0 < idx < 9:
        return None
    return commons.REQ.get(f"https://it.kegs.org.uk/wp-content/themes/corporate-v3/headers/header_{idx}.jpg").content


def download_banner():
    return commons.REQ.get("https://it.kegs.org.uk/wp-content/themes/corporate-v3/images/banner-page.jpg").content



"""
Session class and login function. (most bromcom functionality)
"""

from __future__ import annotations

from base64 import b64decode

import dateparser
import requests
import mimetypes
from datetime import datetime, timedelta

from bs4 import BeautifulSoup, SoupStrainer

from ..util import exceptions, commons

import atexit

from . import timetable


class Session:
    def __init__(self, _sess: requests.Session):
        self._sess: requests.Session = _sess

        self._name: str | None = None

        self._timetable_weeks: list[timetable.WeekDate] | None = None

        atexit.register(self.logout)

    def __repr__(self):
        return f"Session for {self.name}"

    def logout(self) -> requests.Response:
        """
        Send a logout request to bromcom. After this is called, the session will no longer function.
        :return: The response from bromcom
        """
        resp = self._sess.get("https://www.bromcomvle.com/Auth/Logout")
        print(f"Logged out with status code: {resp.status_code}")
        self._sess = None

        return resp

    # --- Account settings ---
    def set_color_preference(self, *, name: str = "Theme", value: str = "default"):
        """
        Set a color preference request to bromcom. Might not work yet
        """
        return self._sess.post("https://www.bromcomvle.com/AccountSettings/SaveColorPreference",
                               json={
                                   "Name": name,
                                   "Value": value
                               })

    @property
    def email(self):
        """
        Fetch the user email from the account settings page
        """
        text = self._sess.get("https://www.bromcomvle.com/AccountSettings").text
        soup = BeautifulSoup(text, "html.parser")

        email_inp = soup.find("input", {
            "class": "form-control",
            "id": "EmailAddress",
            "name": "EmailAddress"
        })

        return email_inp["value"]

    @property
    def school_contact_details(self) -> dict:
        """
        Fetch the school contact details as a key:value table from the hidden drop-down menu
        """
        text = self._sess.get("https://www.bromcomvle.com/Home/Dashboard").text
        soup = BeautifulSoup(text, "html.parser")

        conn_anchor = soup.find("a", {"title": "Contact School"})
        table = conn_anchor.parent.find("table")

        data = {}
        for tr in table.find_all("tr"):
            tr_data = []
            for i, td in enumerate(tr.find_all("td")):
                text: str = td.text
                if i == 0:
                    continue

                elif text.endswith(':'):
                    # Trim off colon
                    text = text[:-1]

                tr_data.append(text)

            if len(tr_data) == 2:
                # Only add stuff that can be made into a dict
                data[tr_data[0]] = tr_data[1]

        return data

    @property
    def name(self):
        """
        Fetch the student name (not username) from the deshboard page
        """
        if self._name is None:
            text = self._sess.get("https://www.bromcomvle.com/Home/Dashboard").text
            soup = BeautifulSoup(text, "html.parser", parse_only=SoupStrainer("span"))

            message = soup.find("span", {"id": "WelcomeMessage"})
            if message is None:
                raise exceptions.NotFound(f"Could not find welcome message! Response: {text}")

            self._name = commons.webscrape_section(message.text, "Hi ", ". Welcome Back!")

        return self._name

    @property
    def pfp(self) -> bytes:
        """
        Fetch the user's corresponding profile picture as bytes
        """
        return self._sess.get("https://www.bromcomvle.com/AccountSettings/GetPersonPhoto").content

    @property
    def school_photo(self) -> bytes:
        """
        Fetch the school's corresponding photo as bytes
        """
        return self._sess.get("https://www.bromcomvle.com/AccountSettings/GetSchoolPhoto").content

    @property
    def pfp_ext(self):
        """
        Fetch the image format of the profile picture
        """
        response = self._sess.get("https://www.bromcomvle.com/AccountSettings/GetPersonPhoto")
        return mimetypes.guess_extension(response.headers.get("Content-Type", "image/Jpeg"))

    @property
    def school_photo_ext(self):
        """
        Fetch the image format of the school picture
        """
        response = self._sess.get("https://www.bromcomvle.com/AccountSettings/GetSchoolPhoto")
        return mimetypes.guess_extension(response.headers.get("Content-Type", "image/Jpeg"))

    # --- Timetable methods ---
    def get_timetable_list(self, start_date: datetime | timetable.WeekDate = None,
                           end_date: datetime | timetable.WeekDate = None, w_a_b: str = None) -> list[
        timetable.Lesson]:
        """
        Fetch the user's timetable starting at a corresponding week and ending on another, as a list of Lesson objects
        :param w_a_b: week a or b on start date
        :param start_date: The start date given to bromcom. Can be a datetime or a WeekDate object. Defaults to the latest valid week.
        :param end_date: The end date fiven to bromcom. Defaults to a week ahead of the start date.
        :return: A list of lesson objects, each with a period #, subject name, class name, room name etc.
        """

        if isinstance(start_date, timetable.WeekDate):
            start_date = start_date.date
        if isinstance(end_date, timetable.WeekDate):
            end_date = end_date.date

        if start_date is None:
            start_date = self.current_week.date
        if end_date is None:
            end_date = start_date + timedelta(weeks=1)

        if w_a_b is None:
            w_a_b = self.timetable_weeks.index(self.get_tt_week(start_date)) % 2
        else:
            w_a_b = "ab".index(w_a_b)

        response = self._sess.get("https://www.bromcomvle.com/Timetable/GetTimeTable",
                                  params={
                                      "WeekStartDate": commons.to_dformat(start_date),
                                      "weekEndDate": commons.to_dformat(end_date),
                                      "type": 1
                                  })
        data = response.json()["table"]

        lessons = []
        for lesson_data in data:
            lesson_data: dict[str, str | int]

            lesson_start_date = datetime.fromisoformat(
                        lesson_data.get("startDate")
                    )
            lessons.append(
                timetable.Lesson(
                    lesson_data.get("periods"),
                    lesson_data.get("subject"),
                    lesson_data.get("class"),
                    lesson_data.get("room"),
                    lesson_data.get("teacherName"),
                    lesson_data.get("teacherID"),
                    lesson_data.get("weekID"),
                    lesson_start_date,
                    datetime.fromisoformat(
                        lesson_data.get("endDate")
                    ),
                    color=lesson_data.get("subjectColour"),
                    _sess=self,
                    week_a_b="ab"[(self.get_tt_week_idx(lesson_start_date) - self.get_tt_week_idx(start_date) + w_a_b) % 2]
                ))
        return lessons

    def get_weeks_a_b(self, delta: int = 5):
        """
        :param delta: number of weeks before and after to measure
        :return:
        """
        idx = self.current_week_idx

        weeks0 = []
        weeks1 = []

        for i in range(idx - delta, idx + delta):
            if not 0 <= i < len(self.timetable_weeks):
                continue

            week = self.timetable_weeks[i]
            _timetable = self.get_timetable_list(week)

            for lesson in _timetable:
                if lesson.week_a_b == "a":
                    weeks0.append(lesson)
                else:
                    weeks1.append(lesson)

        return weeks0, weeks1

    def get_mode_timetables(self, delta: int = 5):
        """
        Infer the base timetable. Look over multiple weeks to avoid being tripped up by pshe, or Comp Room lessons etc.
        :param delta: # of weeks forward/back to look at
        :return: a dictionary of a dictionary of a dictionary of lessons
        """
        a, b = self.get_weeks_a_b(delta)
        return {"a": timetable.get_mode_timetable(a),
                "b": timetable.get_mode_timetable(b)}

    @property
    def timetable_weeks(self) -> list[timetable.WeekDate]:
        """
        Fetch a list of valid weeks in the user's timetable
        :return: A list of WeekDate objects, representing the start of each week, also containing a term and week index.
        """
        if self._timetable_weeks is None:
            self._timetable_weeks = []

            text = self._sess.get("https://www.bromcomvle.com/Timetable").text
            soup = BeautifulSoup(text, "html.parser")

            date_selector = soup.find("select", {"id": "WeekStartDate"})

            for option in date_selector.find_all("option"):
                value = dateparser.parse(option.attrs.get("value"))
                text = option.text

                term, week, _ = text.split(' - ')
                term = commons.webscrape_section(term, "Term ", '', cls=int)
                week = commons.webscrape_section(week, "Week ", '', cls=int)

                self._timetable_weeks.append(
                    timetable.WeekDate(term, week, value)
                )

        return self._timetable_weeks

    def get_tt_week(self, _dtime: datetime) -> timetable.WeekDate | None:
        """
        Gets the timetable week by datetime
        """
        prev = None
        for wdate in self.timetable_weeks:
            if wdate.date > _dtime:
                return prev
            prev = wdate

    @property
    def current_week(self) -> timetable.WeekDate | None:
        """
        Gets the current existing timetable week (will go to last school week during holidays)
        """
        return self.get_tt_week(datetime.today())

    def get_tt_week_idx(self, _dtime: datetime) -> int:
        """
        Gets the timetable week index by datetime
        """
        for i, wdate in enumerate(self.timetable_weeks):
            if wdate.date > _dtime:
                return i
        return -1

    @property
    def current_week_idx(self) -> int:
        return self.get_tt_week_idx(datetime.today())

    # --- Attendance methods ---
    @property
    def present_late_ratio(self) -> dict[str, int]:
        """
        Webscrape JSON inside JS inside HTML to get the present, late and other attendance type counts. i.e.:
        Returns a dictionary e.g.: {
        "present": 176,
        "late": 21
        }
        :return: A dictionary of attendance statuses and their counts
        """
        # Parse JSON inside of JS inside of HTML. Yeah....
        text = self._sess.get("https://www.bromcomvle.com/Attendance").text
        soup = BeautifulSoup(text, "html.parser")

        script_prf = ('$(document).ready(function () {\r\n'
                      '        var AttendanceChart = c3.generate({\r\n'
                      '            bindto: \'#AttendanceChart\',\r\n'
                      '            data: {\r\n'
                      '                columns: ')

        for script in soup.find_all("script", {"type": "text/javascript"}):
            text = script.text.strip()

            if text.startswith(script_prf):
                # Found correct js script. Now webscrape.
                text = text[len(script_prf):]

                data = commons.consume_json(text)

                ret = {cat: count for cat, count in data}

                return ret

        return {}

    @property
    def attendance_status(self):
        """
        Get the Status for the current day. (Uses the widget api)
        """
        return self._sess.get("https://www.bromcomvle.com/Home/GetAttendanceWidgetData").json()

    # --- Reports data ---

    @property
    def reports_data(self) -> dict[str, list[dict[str, str]]]:
        """
        Fetch the report list (needs to be parsed)
        :return: A list of dictionaries representing reports. The filePath attribute can be used in the get_report method to fetch the report pdf as bytes
        """
        # Parse this later
        return self._sess.get("https://www.bromcomvle.com/Home/GetReportsWidgetData").json()

    def get_report(self, filepath: str) -> bytes:
        """
        Get the report with the given 'filepath' as bytes
        :param filepath: The filePath attribute in the report data
        :return: The report data as bytes
        """
        # Get the data encoded in b64 encoded in JSON. Weird.
        data = self._sess.get("https://www.bromcomvle.com/Report/GetReport",
                              params={
                                  "filePath": filepath
                              }).json()

        return b64decode(data)

    # --- Exam data ---

    @property
    def exam_data(self) -> list[dict[str, str]]:
        """
        Fetch the exam data from the widget api
        :return:
        """
        # Parse this
        return self._sess.get("https://www.bromcomvle.com/Home/GetExamResultsWidgetData").json()

    # --- Bookmarks data ---
    @property
    def bookmarks_data(self) -> list[dict]:
        """
        Get the bookmarks list as a list of dictionaries (needs to be parsed)
        :return: list of dictionaries, each is a bookmark
        """
        # Parse this
        return self._sess.get("https://www.bromcomvle.com/Home/GetBookmarksWidgetData").json()

    # --- Homework data ---
    @property
    def homework_data(self) -> list:
        """
        Fetch homework data using the widget api. I have no homework so I am unable to parse this
        :return: A list of something
        """
        return self._sess.get("https://www.bromcomvle.com/Home/GetHomeworkWidgetData").json()


def login(school_id: int, username: str, password: str, remember_me: bool = True) -> Session:
    """
    Login to bromcom with a school id, username and password.
    :param school_id: KEGS school id (you provide it)
    :param username: Your username
    :param password: Your password
    :param remember_me: Option to 'remember me.' Defaults to True
    :return: A session representing your login
    """
    _sess = requests.Session()
    _sess.headers = commons.headers.copy()

    text = _sess.get("https://www.bromcomvle.com/").text
    soup = BeautifulSoup(text, "html.parser", parse_only=SoupStrainer("input"))

    rvinp = soup.find("input", {"name": "__RequestVerificationToken"})
    if rvinp is None:
        ptfy = BeautifulSoup(text, "html.parser").prettify()
        raise exceptions.NotFound(f"Could not find rv token; response text: {ptfy}")

    rvtoken = rvinp.attrs.get("value")

    response = _sess.post("https://www.bromcomvle.com/",
                          data={
                              "SpaceID": '',

                              "schoolid": school_id,
                              "username": username,
                              "password": password,

                              "__RequestVerificationToken": rvtoken,
                              "rememberme": str(remember_me).lower()
                          })

    if response.status_code != 200:
        if response.status_code == 500:
            raise exceptions.ServerError(
                f"The bromcom server experienced some error when handling the login request (ERR 500). Response content: {response.content}")
        else:
            raise exceptions.Unauthorised(
                f"The provided details for {username} may be invalid. Status code: {response.status_code} "
                f"Response content: {response.content}")

    return Session(_sess)



"""
The ultimate KEGS webscraping experience.
"""
from __future__ import annotations

from . import bromcom, it, oliver, papercut, site, util, vle, kerboodle, activelearn, ms365
from .util import exceptions



"""
Anything to do with the main website: https://kegs.org.uk/

Should really only be about fetching data, not posting any
"""
from .asset import download_asset_by_id, find_asset_ids
from .events import get_events_pdf, get_calendar_page
from .event_pdf_reader import read as read_event_pdf



"""
Session class and login/login by moodle function
"""
from __future__ import annotations

import json
import re
import atexit
import warnings

from typing import Literal, Any
from datetime import datetime
from dataclasses import dataclass, field

import dateparser
import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse, parse_qs

from . import file, user, forum, blog, tag, calendar, course
from ..util import commons, exceptions


@dataclass
class Session:
    """
    Represents a login session
    """
    _sesskey: str | None = None
    _file_client_id: str | None = None
    _file_item_id: str | None = None
    _user_id: int | None = None
    _user: user.User | None = None
    _username: str | None = None

    rq: requests.Session = None

    def __post_init__(self):
        """Request handler (requests session object)"""

        self.assert_login()

        atexit.register(self.logout)

    # --- Session/auth related methods ---
    def unregister_logout(self):
        atexit.unregister(self.logout)

    @property
    def moodlesession(self):
        return self.rq.cookies.get("MoodleSession")

    @property
    def is_signed_in(self):
        return self.rq.get("https://vle.kegs.org.uk").url != "https://vle.kegs.org.uk/login/index.php"

    @property
    def sesskey(self):
        """Get the sesskey query parameter used in various functions. Webscraped from JS..."""
        if self._sesskey is None:
            pfx = "var M = {}; M.yui = {};\nM.pageloadstarttime = new Date();\nM.cfg = "

            response = self.rq.get("https://vle.kegs.org.uk/")
            soup = BeautifulSoup(response.text, "html.parser")

            self._sesskey = None
            for script in soup.find_all("script"):
                text = script.text
                if "\"sesskey\":" in text:
                    i = text.find(pfx)
                    if i > -1:
                        i += len(pfx) - 1
                        data = commons.consume_json(text, i)

                        if isinstance(data, dict):
                            self._sesskey = data.get("sesskey")

        return self._sesskey

    def connect_notifications(self, *, limit: int = 20, offset: int = 0,
                              user_id: int = None, newestfirst: bool = True) -> tuple[int, list[dict[str, Any]]]:
        """
        Because KEGSNet messaging is disabled, this is mostly useless
        I can still work out what would be the response format from the moodle docs, but no point
        """
        if user_id is None:
            user_id = self.user_id

        data = self.webservice("message_popup_get_popup_notifications",
                               limit=limit, offset=offset, useridto=user_id, newestfirst=int(newestfirst))

        return data["unreadcount"], data["notifications"]

    @property
    def file_client_id(self):
        """Get the client id value used for file management"""
        if self._file_client_id is None:
            response = self.rq.get("https://vle.kegs.org.uk/user/files.php")
            soup = BeautifulSoup(response.text, "html.parser")

            for div in soup.find_all("div", {"class": "filemanager w-100 fm-loading"}):
                self._file_client_id = div.attrs["id"].split("filemanager-")[1]

        return self._file_client_id

    @property
    def connected_user(self) -> user.User:
        """Fetch the connected user to this session"""
        if not self._user:
            self._user = self.connect_user_by_id(self.user_id)

        return self._user

    @property
    def file_item_id(self):
        """Fetch the item id value used for file management"""
        if self._file_item_id is None:
            response = self.rq.get("https://vle.kegs.org.uk/user/files.php")
            soup = BeautifulSoup(response.text, "html.parser")
            self._file_item_id = soup.find("input", {"id": "id_files_filemanager"}).attrs.get("value")

        return self._file_item_id

    @property
    def username(self):
        """Fetch the connected user's username"""
        if self._username is None:
            response = self.rq.get("https://vle.kegs.org.uk/login/index.php")
            soup = BeautifulSoup(response.text, "html.parser")
            for alert_elem in soup.find_all(attrs={"role": "alert"}):
                alert = alert_elem.text

                username = commons.webscrape_value(alert, "You are already logged in as ",
                                                   ", you need to log out before logging in as different user.")
                if username:
                    self._username = username
                    break

        return self._username

    @property
    def user_id(self):
        """Fetch the connected user's user id"""
        if self._user_id is None:
            response = self.rq.get("https://vle.kegs.org.uk/")
            soup = BeautifulSoup(response.text, "html.parser")

            url = soup.find("a", {"title": "View profile"}) \
                .attrs["href"]

            parsed = parse_qs(urlparse(url).query)
            self._user_id = int(parsed.get("id")[0])

        return self._user_id

    def assert_login(self):
        """Raise an error if there is no connected user"""
        assert self.is_signed_in

    def logout(self):
        """
        Send a logout request to KEGSNet. After this is called, the session is supposed to no longer function.
        :return: The response from KEGSNet
        """
        response = self.rq.get("https://vle.kegs.org.uk/login/logout.php",
                               params={"sesskey": self.sesskey})
        print(f"Logged out with status code {response.status_code}")
        return response

    # --- Connecting ---
    def connect_user_by_id(self, _id: int) -> user.User:
        """Get a user by ID and attach this session object to it"""
        ret = user.User(_id, _session=self)
        ret.update_from_id()
        return ret

    def get_users(self, _id: int | list[int] | None = None, idnumber: int | list[int] | None = None, username: str | list[str] | None = None, email: str | list[str] | None = None):
        if _id:
            _field = "id"
            _value = _id
        elif idnumber:
            _field = "idnumber"
            _value = idnumber
        elif username:
            _field = "username"
            _value = username
        elif email:
            _field = "email"
            _value = email
        else:
            raise ValueError("Nothing to search by")

        if not isinstance(_value, list):
            _value = [_value]

        data = self.webservice("core_user_get_users_by_field", field=_field, values=_value)
        return data

    def connect_partial_user(self, **kwargs):
        """
        Connect to a user with given kwargs without any updating
        """
        return user.User(_session=self, **kwargs)

    def connect_forum(self, _id: int) -> forum.Forum:
        """Get a forum by ID and attach this session object to it"""
        ret = forum.Forum(_id, _session=self)
        ret.update_by_id()
        return ret

    def connect_site_news(self):
        return self.connect_forum(377)

    # --- Private Files ---
    def _file_data(self, fp: str) -> dict:
        """Fetch the JSON response for private files in a given directory"""

        # Believe or not, KegsNet does actually have some JSON endpoints!
        return self.rq.post("https://vle.kegs.org.uk/repository/draftfiles_ajax.php",
                            params={"action": "list"},
                            data={
                                "sesskey": self.sesskey,

                                "clientid": self.file_client_id,
                                "itemid": self.file_item_id,
                                "filepath": fp
                            }).json()

    def files_in_dir(self, fp: str) -> list[file.File]:
        """Fetch files in a given directory"""
        data = self._file_data(fp)["list"]
        files = []
        for file_data in data:
            files.append(file.File.from_json(file_data, self))
        return files

    @property
    def files(self):
        """Fetch the files in the root directory"""
        return self.files_in_dir('/')

    def add_filepath(self, fp: str, data: bytes, author: str = '', _license: str = "unknown"):
        """
        Add file by path - infer file title and file path, e.g. foo/bar/baz.txt
        :param fp:
        :param data:
        :param author:
        :param _license:
        :return:
        """
        split = fp.split('/')
        fp = '/'.join(split[:-1])
        self.add_file(split[-1], data, author, _license, fp)

    def add_file(self, title: str, data: bytes, author: str = '', _license: str = "unknown", fp: str = '/',
                 save_changes: bool = True):
        """
        Make a POST request to add a new file to the given filepath


        NOTE: KEGSNet automatically removes slashes from the title. if you want to put the file in a subdirectory, add slashes in the `fp` parameter

        If the filename already exists, KEGSNet will automatically add a number on. e.g. foo.txt -> foo (1).txt

        :param title: file title
        :param data: file content (bytes)
        :param author: Author metadata. Defaults to ''
        :param _license: Given license. Defaults to 'unknown'
        :param fp: Directory path to add the file. Defaults to the root directory
        :param save_changes: Whether to save the change. Defaults to True
        """
        # Perhaps this method should take in a File object instead of title/data/author etc

        self.rq.post("https://vle.kegs.org.uk/repository/repository_ajax.php",
                     params={"action": "upload"},
                     data={
                         "sesskey": self.sesskey,
                         "repo_id": 3,  # I'm not sure if it has to be 3

                         "title": title,
                         "author": author,
                         "license": _license,

                         "clientid": self.file_client_id,
                         "itemid": self.file_item_id,
                         "savepath": fp
                     },
                     files={"repo_upload_file": data})

        # Save changes
        if save_changes:
            self.file_save_changes()

    def file_save_changes(self):
        """
        Tell kegsnet to save our changes to our files
        """
        self.rq.post("https://vle.kegs.org.uk/user/files.php",
                     data={"returnurl": "https://vle.kegs.org.uk/user/files.php",

                           "sesskey": self.sesskey,
                           "files_filemanager": self.file_item_id,
                           "_qf__user_files_form": 1,
                           "submitbutton": "Save changes"})

    @property
    def file_zip(self) -> bytes:
        """
        Returns bytes of your files as a zip archive
        """
        url = self.rq.post("https://vle.kegs.org.uk/repository/draftfiles_ajax.php",
                           params={"action": "downloaddir"},
                           data={
                               "sesskey": self.sesskey,
                               "client_id": self.file_client_id,
                               "filepath": '/',
                               "itemid": self.file_item_id
                           }).json()["fileurl"]

        return self.rq.get(url).content

    # --- Blogs ---
    def _find_blog_entires(self, soup: BeautifulSoup) -> list[blog.Entry]:
        entries = []
        for div in soup.find("div", {"role": "main"}).find_all("div"):
            raw_id = div.attrs.get("id", '')

            if re.match(r"b\d*", raw_id):
                entries.append(blog.Entry(_session=self))
                entries[-1].update_from_div(div)

        return entries

    def connect_user_blog_entries(self, userid: int = None, *, limit: int = 10, offset: int = 0) -> list[blog.Entry]:
        warnings.warn("This will be deprecated soon. Try to use connect_blog_entries instead")
        if userid is None:
            userid = self.user_id

        entries = []
        for page, _ in zip(*commons.generate_page_range(limit, offset, 10, 0)):
            text = self.rq.get("https://vle.kegs.org.uk/blog/index.php",
                               params={
                                   "blogpage": page,
                                   "userid": userid
                               }).text
            soup = BeautifulSoup(text, "html.parser")
            entries += self._find_blog_entires(soup)

        return entries

    def connect_blog_entries(self, *, limit: int = 10, offset: int = 0,
                             # search filters
                             _tag: tag.Tag = None,
                             _course: course.Course = None,
                             _user: user.User = None,
                             tagname: str = None,
                             tagid: int = None,
                             userid: int = None,
                             cmid: int = None,  # idk what this one is
                             entryid: int = None,
                             groupid: int = None,
                             courseid: int = None,
                             search: str = None,
                             ):
        if offset != 0:
            warnings.warn("offset+limit -> page+perpage conversion has not been made yet! Offset will be ignored",
                          category=exceptions.UnimplementedWarning)

        filters = []

        def add_filter(name: str, value):
            if value is not None:
                filters.append({"name": name, "value": value})

        if _tag:
            tagid = _tag.id
        if _user:
            userid = _user.id
        if _course:
            courseid = _course.id

        add_filter("tag", tagname)
        add_filter("tagid", tagid)
        add_filter("userid", userid)
        add_filter("cmid", cmid)
        add_filter("entryid", entryid)
        add_filter("groupid", groupid)
        add_filter("courseid", courseid)
        add_filter("search", search)

        data = self.webservice("core_blog_get_entries",
                               page=0, perpage=limit, filters=filters)

        return [blog.Entry.from_json(entry_data, self) for entry_data in data["entries"]]

    def connect_blog_entry_by_id(self, _id: int):
        entry = blog.Entry(id=_id, _session=self)
        entry.update_from_id()
        return entry

    # --- Tags ---

    def connect_tag_by_name(self, name: str) -> tag.Tag:
        _tag = tag.Tag(name, _session=self)
        _tag.update()
        return _tag

    def connect_tag_by_id(self, _id: int) -> tag.Tag:
        _tag = tag.Tag(id=_id, _session=self)
        _tag.update()
        return _tag

    # --- Calendar ---

    def connect_calendar(self, view_type: Literal["month", "day", "upcoming"] = "day",
                         _time: int | float | datetime = None, _course: int | str | course.Course = None):
        if isinstance(_time, datetime):
            _time = _time.timestamp()

        if isinstance(_course, course.Course):
            _course = _course.id

        resp = self.rq.get("https://vle.kegs.org.uk/calendar/view.php",
                           params={
                               "view": view_type,
                               "time": _time,
                               "course": _course
                           })
        ret = calendar.Calendar(_sess=self)

        soup = BeautifulSoup(resp.text, "html.parser")
        div = soup.find("div", {"class": "calendarwrapper"})

        if view_type == "month":
            ...
        elif view_type in ("day", "upcoming"):
            evlist = div.find("div", {"class": "eventlist"})
            for event_div in evlist.find_all("div", {"data-type": "event"}):
                cal_event = calendar.Event(_sess=self)

                head = event_div.find("div", {"class": "box card-header clearfix calendar_event_user"})

                cal_event.title = head.text.strip()

                body = event_div.find("div", {"class": "description card-body"})

                for row in body.find_all("div", {"class": "row"}):
                    row_type = row.find("i").get("title")

                    row_text = row.text.strip()
                    match row_type:
                        case "When":
                            cal_event.date = dateparser.parse(row_text)

                        case "Event type":
                            cal_event.type = row_text

                        case "Location":
                            cal_event.location = row_text

                        case "Description":
                            cal_event.description = row_text

                        case _:
                            warnings.warn(
                                f"Did not recognise calendar row type: {row_type!r} - report this on github: https://github.com/BigPotatoPizzaHey/kegscraper")
                ret.events.append(cal_event)

        return ret

    # -- Courses -- #
    def connect_recent_courses(self, limit: int = 10, offset: int = 0):
        data = self.webservice("core_course_get_recent_courses", limit=limit, offset=offset)
        return [course.Course.from_json(course_data) for course_data in data]

    def webservice(self, name, /, **args):
        """
        Directly interact with the webservice api
        :param name:methodname of webservice api, e.g. core_course_search_courses
        :param args:args to send to webservice api
        :return:
        """
        data: list = self.rq.post("https://vle.kegs.org.uk/lib/ajax/service.php",
                     params={"sesskey": self.sesskey},  # "info": name
                     json=[{"methodname": name, "args": args}]).json()

        skip = False
        if isinstance(data, dict):
            skip = "error" in data

        if not skip:
            data: dict[str, dict | str | int | float | None | bool | list] = data[0]

        if data["error"]:
            try:
                raise exceptions.WebServiceError(
                    f"{data['exception']['errorcode']!r}: {data['exception']['message']!r}")
            except KeyError:
                try:
                    raise exceptions.WebServiceError(
                        f"{data['errorcode']!r}: {data['error']!r}")
                except KeyError:
                    raise exceptions.WebServiceError(f"Error: {data}")

        return data["data"]

    def search_courses(self, query: str):
        data = self.webservice("core_course_search_courses", criterianame="tagid", criteriavalue=query)
        return data

    def connect_enrolled_courses(self, classification: Literal["future", "inprogress", "past"] = "inprogress",
                                 limit: int = 9999, offset: int = 0):
        data = self.webservice("core_course_get_enrolled_courses_by_timeline_classification",
                               classification=classification, limit=limit, offset=offset)
        return data


# --- * ---

def login(username: str, password: str) -> Session:
    """
    Login to kegsnet with a username and password
    :param username: Your username. Same as your email without '@kegs.org.uk'
    :param password: Your email password
    :return: a new session
    """

    session = requests.Session()
    session.headers = commons.headers.copy()

    response = session.get("https://vle.kegs.org.uk/login/index.php")

    soup = BeautifulSoup(response.text, "html.parser")
    login_token = soup.find("input", {"name": "logintoken"})["value"]

    session.post("https://vle.kegs.org.uk/login/index.php",
                 data={"logintoken": login_token,
                       "anchor": None,
                       "username": username,
                       "password": password
                       })

    return Session(rq=session)


def login_by_moodle(moodle_cookie: str) -> Session:
    """
    Login to kegsnet with just a moodle cookie (basically a session id)
    :param moodle_cookie: The MoodleSession cookie (see in the application/storage tab of your browser devtools when you log in)
    :return: A new session
    """
    session = requests.Session()
    session.cookies.set("MoodleSession", moodle_cookie)

    try:
        return Session(rq=session)
    except requests.exceptions.TooManyRedirects:
        raise ValueError(f"The moodle cookie {moodle_cookie!r} may be invalid/outdated.")



import json

import requests
from datetime import datetime
from dataclasses import dataclass, field
import dateparser

from typing_extensions import deprecated

from bs4 import BeautifulSoup, SoupStrainer
from ..util import commons


@dataclass
class ATCEvent:
    # stands for 'Add-to-calendar' event
    """
    Represents a calendar event managed by the atc api
    """
    start: datetime
    end: datetime

    timezone: str = None  # Maybe there is a way to parse this? (e.g. 'Europe/London')

    title: str = None
    description: str = None
    location: str = None
    organiser: str = None

    all_day: bool = False
    """Attribute storing if it is an all-day event for compatability with the KEGS Events PDF"""

    @classmethod
    def from_data(cls, data: dict[str, str]):
        """Load an event from a dictionary of data from the atc api"""
        return cls(
            dateparser.parse(data.get("atc_date_start")),
            dateparser.parse(data.get("atc_date_end")),

            data.get("atc_timezone"),
            data.get("atc_title"),
            data.get("atc_description"),
            data.get("atc_location"),
            data.get("atc_organizer"),
        )

    def __repr__(self):
        return f"Event<{self.title!r} {commons.to_dformat(self.start, sep='/')} - {commons.to_dformat(self.end, sep='/')}>"


@dataclass
class CalendarEvent:
    """
    Represents a calendar event stored on the kegs main website
    """
    id: int

    code_name: str = '_'  # Used for fetching url, but doesn't actually matter

    _event_data: ATCEvent = field(repr=False, default=None)

    @property
    def url(self) -> str:
        """Get a link to the calendar event. This will time out at some point"""
        return f"https://kegs.org.uk/{self.code_name}/{self.id}.html"

    @property
    def event_data(self) -> ATCEvent:
        """Get the actual data for the calendar event, and return as an ATCEvent object"""
        if self._event_data is None:
            text = commons.REQ.get(self.url).text
            soup = BeautifulSoup(text, "html.parser")

            atc_btn = soup.find("span", {"class": "addtocalendar"})
            event_var = atc_btn.find("var", {"class": "atc_event"})

            data = {}
            for var in event_var.find_all("var"):
                data[' '.join(var["class"])] = var.text

            self._event_data = ATCEvent.from_data(data)

        return self._event_data

@deprecated("This endpoint has been removed")
def get_events_pdf() -> bytes:
    """
    Fetch the KEGS event list pdf as bytes
    """
    return commons.REQ.get("https://www.kegs.org.uk/eventsPDF.cfm").content

@deprecated("This endpoint has been removed")
def get_calendar_page(date: datetime | str = None, by: str = "MONTH") -> list[CalendarEvent]:
    """
    Fetch the list of calendar events corresponding the page on the calendar
    :param date: Date to fetch data around
    :param by: Event fetching mode: either 'WEEK' or 'MONTH'
    :return: A list of CalendarEvent objects
    """
    if date is None:
        date = datetime.today()

    if isinstance(date, datetime):
        # Maybe edit commons.to_dformat
        date = f"{str(date.day).zfill(2)}/{str(date.month).zfill(2)}/{date.year}"

    by = by.upper()
    response = commons.REQ.get("https://www.kegs.org.uk/plugins/cfc/bycontentid/website/calendar.cfc",
                            params={
                                "method": "fncDisplayMobile",
                                "returnFormat": "json",  # It doesn't return JSON though...

                                "_cf_ajaxproxytoken": "670A99ACD5D7DDD0",  # This seems to be consistent across sessions
                                "_cf_clientid": "B08044EF38BAB77125840DD9A5ADFCDF",  # Also consistent

                                "_cf_nodebug": True,
                                "_cf_nocache": True,
                                "_cf_rc": 0,

                                "argumentCollection": json.dumps({
                                    "mobileMode": by,
                                    "mobileDate": date,
                                    "mobileHide": 0,
                                    "mobileCat": 2693,
                                    "mobileSearch": "search",
                                    "mobileType": "all",
                                    "mobileDept": "all"
                                })
                            })

    soup = BeautifulSoup(response.text, "html.parser", parse_only=SoupStrainer("a"))

    partial_events = []
    for anchor in soup.find_all("a"):
        if "href" in anchor.attrs:
            href = '"' + anchor["href"] + '"'
            href: str = json.loads(json.loads(href))  # Yeah, it seems to be double json encoded

            _, name, cid = href.split('/')

            cid = commons.keep_chrs(cid, cls=int)

            partial_events.append(CalendarEvent(
                cid, code_name=name
            ))

    return partial_events



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from datetime import datetime
from urllib.parse import urlparse, parse_qs
from typing import Any, Self

import dateparser
from bs4 import PageElement, BeautifulSoup

from . import session, user, tag, file
from ..util import commons, exceptions


@dataclass
class External:
    """Represents an external blog or an external blog entry"""
    url: str = None
    name: str = None
    id: int = None

    _session: session.Session = field(repr=False, default=None)


@dataclass
class Comment:
    id: int = None
    content: BeautifulSoup = field(repr=False, default=None)
    format: str = '0'  # Idk what this is
    created: datetime = None
    author: user.User = None
    deletable: bool = False

    _entry: Entry = field(repr=False, default=None)
    _session: session.Session = field(repr=False, default=None)

    @classmethod
    def from_json(cls, data: dict[str, str | bool], _entry: Entry, _session: session.Session) -> Self:
        return cls(
            data.get("id"),
            BeautifulSoup(data.get("content", ''), "html.parser"),
            data.get("format"),
            datetime.fromtimestamp(int(data.get("timecreated"))),
            _session.connect_partial_user(
                id=int(data.get("userid")),
                name=data.get("fullname")
            ),
            data.get("delete"),
            _entry,
            _session
        )

    @property
    def text(self):
        return self.content.text

    def delete(self):
        response =  self._session.rq.post("https://vle.kegs.org.uk/comment/comment_ajax.php",
                              data={
                                  "sesskey": self._session.sesskey,
                                  "action": "delete",
                                  "client_id": self._session.file_client_id,
                                  "itemid": self._entry.id,
                                  "area": "format_blog",
                                  "courseid": 1,
                                  "contextid": self._entry.context_id,
                                  "component": "blog",
                                  "commentid": self.id
                              })
        if response:
            data = response.json()
            if data.get("error") == "Invalid comment ID":
                raise exceptions.NotFound("Invalud comment ID")
            else:
                return data
        else:
            extra = f" This may be because you are not {self.author.name} ({self.author.id})" if self._session.user_id != self.author.id else ''
            warnings.warn(f"Possibly couldn't delete {self}.{extra}")


@dataclass
class Entry:
    id: int = None
    subject: str = None
    author: user.User = None
    date_created: datetime = None
    date_modified: datetime = None
    publishstate: str = None
    content: PageElement | Any = field(repr=False, default=None)

    attachments: list[file.File] = None
    images: PageElement | Any = field(repr=False, default=None)
    tags: list[tag.Tag] = None

    external_blog: External = None
    external_blog_entry: External = None

    context_id: int = None

    _session: session.Session = field(repr=False, default=None)

    @property
    def url(self):
        return f"https://vle.kegs.org.uk/blog/index.php?entryid={self.id}"

    @classmethod
    def from_json(cls, data: dict[str | Any], _sess: session.Session) -> Self:
        if data["module"] == "blog_external":
            ext_be = External(url=data["uniquehash"], _session=_sess)
            ext = External(id=data["moduleid"], _session=_sess)
        else:
            ext_be = None
            ext = None

        return cls(
            _session=_sess,
            id=data["id"],
            subject=data["subject"],
            author=user.User(id=data["userid"], _session=_sess),
            date_created=datetime.fromtimestamp(data["created"]),
            date_modified=datetime.fromtimestamp(data["lastmodified"]),
            publishstate=data["publishstate"],
            content=BeautifulSoup(data["summary"], "html.parser"),
            attachments=[file.File.from_json2(attch, _sess) for attch in data["attachmentfiles"]],
            tags=[tag.Tag.from_json(td, _sess) for td in data["tags"]],
            external_blog=ext,
            external_blog_entry=ext_be,
        )

    def update_from_id(self):
        text = self._session.rq.get("https://vle.kegs.org.uk/blog/index.php",
                                    params={
                                        "entryid": self.id
                                    }).text
        soup = BeautifulSoup(text, "html.parser")

        self.update_from_div(
            soup.find("div", {"id": f"b{self.id}"})
        )

    def update_from_div(self, div: PageElement):
        if div is None:
            raise exceptions.NotFound(
                f"BlogEntry #{self.id}, ({self}) does not seem to exist. It may have been deleted, or may never have existed, or you may be logged out.")

        header = div.find("div", {"class": "row header clearfix"})
        main = div.find("div", {"class": "row maincontent clearfix"})

        if header is None and main is None:
            raise exceptions.NotFound(
                f"BlogEntry #{self.id}, ({self}) does not seem to exist. It may have been deleted, or may never have existed, or you may be logged out.")

        self.id = int(div["id"][1:])

        self.subject = header.find("div", {"class": "subject"}).text

        author_anchor = header.find("div", {"class": "author"}).find("a")
        parse = urlparse(author_anchor["href"])
        qparse = parse_qs(parse.query)

        author_id = int(qparse["id"][0])

        self.author = self._session.connect_partial_user(id=author_id, name=author_anchor.text)

        date_str = author_anchor.next.next.text
        self.date_created = dateparser.parse(date_str)

        external_div = header.find("div", {"class": "externalblog"})
        if external_div:
            external_anchor = external_div.find("a")
            if external_anchor:
                self.external_blog = External(
                    external_anchor["href"],
                    external_anchor.text,
                    _session=self._session
                )

        # Get actual blog content
        audience = main.find("div", {"class": "audience"}).text.strip()
        self.publishstate = {"Anyone on this site": "site"}.get(audience)

        self.images = main.find("div", {"class": "attachedimages"})

        self.content = main.find("div", {"class": "no-overflow"}) \
            .find("div", {"class": "no-overflow"})

        external_div = main.find("div", {"class": "externalblog"})
        if external_div:
            external_anchor = external_div.find("a")
            if external_anchor:
                self.external_blog_entry = External(
                    external_anchor["href"],
                    external_anchor.text,
                    _session=self._session
                )

        tag_list = main.find("div", {"class": "tag_list"})
        if tag_list:
            tag_list = tag_list.find_all("li")

        self.tags = []
        if tag_list:
            for tag_data in tag_list:
                tag_a = tag_data.find("a")

                # We could probably also get the anchor text, but this is more robust
                parse = urlparse(tag_a["href"])
                qparse = parse_qs(parse.query)

                self.tags.append(tag.Tag(qparse["tag"][0], _session=self._session))

        mdl = main.find("div", {"class": "mdl-left"})
        njs_url = mdl.find("a", {"class": "showcommentsnonjs"})["href"]
        parse = urlparse(njs_url)
        qparse = parse_qs(parse.query)
        self.context_id = int(qparse["comment_context"][0])

    def get_comments(self, *, limit: int = 1, offset: int = 0) -> list[Comment]:
        if self.context_id is None:
            self.update_from_id()

        data_lst = []
        for page, _ in zip(*commons.generate_page_range(limit, offset, items_per_page=999, starting_page=0)):
            data_lst += (self._session.rq.post("https://vle.kegs.org.uk/comment/comment_ajax.php",
                                               data={
                                                   "sesskey": self._session.sesskey,
                                                   "action": "get",
                                                   "client_id": self._session.file_client_id,
                                                   "itemid": self.id,
                                                   "area": "format_blog",
                                                   "courseid": "1",
                                                   "contextid": self.context_id,
                                                   "component": "blog",
                                                   "page": page
                                               }).json()["list"])

        return [Comment.from_json(data, self, self._session) for data in data_lst]

    def post_comment(self, content: str) -> Comment:
        if self.context_id is None:
            self.update_from_id()

        response = self._session.rq.post("https://vle.kegs.org.uk/comment/comment_ajax.php",
                                         data={
                                             "sesskey": self._session.sesskey,
                                             "action": "add",
                                             "client_id": self._session.file_client_id,
                                             "itemid": self.id,
                                             "area": "format_blog",
                                             "courseid": 1,
                                             "contextid": self.context_id,
                                             "component": "blog",
                                             "content": content
                                         })

        ret = Comment.from_json(response.json(), self, self._session)
        return ret



"""
Utility functions/variables used commonly across the module
"""

import json
import string
import warnings
import copy
from datetime import datetime
from typing import Any, Final
from inspect import signature

import requests
from bs4 import BeautifulSoup

from . import exceptions

REQ: Final[requests.Session] = requests.Session()

DIGITS: Final = tuple("0123456789")

headers = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36",
    "referer": "https://kegs.org.uk/",
}


def webscrape_value(raw, text_before, text_after, cls: type = str, i1: int = 1, i2: int = 0) -> str | Any:
    return cls(
        raw.split(text_before)[i1] \
            .split(text_after)[i2]
    )


def webscrape_section(raw: str, before: str | int = '', after: str | int = '', cls: type = str) -> str | Any:
    def _toint(val: str | int) -> int:
        return val if isinstance(val, int) else len(val)

    before, after = _toint(before), _toint(after)

    if after == 0:
        section = raw[before:]
    else:
        section = raw[before:-after]
    return cls(section)


def _read_json_number(_string: str) -> float | int:
    ret = ''

    minus = _string[0] == '-'
    if minus:
        ret += '-'
        _string = _string[1:]

    def read_fraction(sub: str):
        sub_ret = ''
        if sub[0] == '.':
            sub_ret += '.'
            sub = sub[1:]
            while sub[0] in DIGITS:
                sub_ret += sub[0]
                sub = sub[1:]

        return sub_ret, sub

    def read_exponent(sub: str):
        sub_ret = ''
        if sub[0].lower() == 'e':
            sub_ret += sub[0]
            sub = sub[1:]

            if sub[0] in "-+":
                sub_ret += sub[0]
                sub = sub[1:]

            if sub[0] not in DIGITS:
                raise exceptions.UnclosedJSONError(f"Invalid exponent {sub}")

            while sub[0] in DIGITS:
                sub_ret += sub[0]
                sub = sub[1:]

        return sub_ret

    if _string[0] == '0':
        ret += '0'
        _string = _string[1:]

    elif _string[0] in DIGITS[1:9]:
        while _string[0] in DIGITS:
            ret += _string[0]
            _string = _string[1:]

    frac, _string = read_fraction(_string)
    ret += frac

    ret += read_exponent(_string)

    return json.loads(ret)


def consume_json(_string: str, i: int = 0) -> str | float | int | dict | list | bool | None:
    """
    Reads a JSON string and stops at the natural end (i.e. when brackets close, or when quotes end, etc.)
    """
    # Named by ChatGPT
    section = _string[i:]
    if section.startswith("true"):
        return True
    elif section.startswith("false"):
        return False
    elif section.startswith("null"):
        return None
    elif section[0] in "0123456789.-":
        return _read_json_number(section)

    depth = 0
    json_text = ''
    out_string = True

    for char in section:
        json_text += char

        if char == '"':
            if len(json_text) > 1:
                unescaped = json_text[-2] != '\\'
            else:
                unescaped = True
            if unescaped:
                out_string ^= True
                if out_string:
                    depth -= 1
                else:
                    depth += 1

        if out_string:
            if char in "[{":
                depth += 1
            elif char in "}]":
                depth -= 1

        if depth == 0 and json_text.strip():
            try:
                return json.loads(json_text.strip())
            except Exception as e:
                warnings.warn(f"Failed to decode: {json_text.strip()!r}")
                raise e

    raise exceptions.UnclosedJSONError(f"Unclosed JSON string, read {json_text}")


def generate_page_range(limit: int, offset: int, items_per_page: int, starting_page: int = 1) -> tuple[
    range, list[int]]:
    """
    Returns a page range (and first indexes per page) generated from a page range
    :param limit: How many items to reach up to
    :param offset: Starting item idx
    :param items_per_page: Number of items per page
    :param starting_page: Page with index '0' (usually 1 actually)
    :return: tuple of page range and list of starting indexes for each page
    """

    if offset < 0:
        raise ValueError(f"offset {offset!r} < 0")
    if limit < 0:
        raise ValueError(f"limit {limit!r} < 0")

    # There are n items on display per page
    # So the first page you need to view is {page 1 idx} + {offset} // {n}

    # The final item to view is at idx {offset} + {limit} - 1
    # (You have to -1 because the index starts at 0)
    # So the page number for this is 1 + (offset + limit - 1) // n

    # But this is a range so we have to add another 1 for the second argument

    page_range = range(
        starting_page + offset // items_per_page,
        1 + starting_page + (offset + limit - 1) // items_per_page
    )
    # start_idxs = [offset] + [0] * (limit // items_per_page)

    return (
        page_range,
        # start_idxs
        [items_per_page * (i - starting_page) for i in page_range]
    )


def find_links(soup: BeautifulSoup) -> list[str]:
    ret = []
    for elem in soup.find_all("a"):
        if "href" in elem.attrs:
            ret.append(str(elem.attrs["href"]))
    return ret


def to_dformat(date: datetime, sep: str = '-') -> str:
    """
    Convert a datetime to the format dd-mm-yyyy
    """

    def i2zs(v: int):
        """convert an integer to a fixed 2-digit string"""
        return str(v).zfill(2)

    return f"{i2zs(date.month)}{sep}{i2zs(date.day)}{sep}{date.year}"


def keep_chrs(_string: str, chars=string.digits, cls: type = str):
    """
    Filter a string to only the characters provided. By default, only keeps digits
    """
    return cls(''.join(filter(
        lambda c: c in chars, _string
    )))


def slice_to_range(slc: slice) -> range:
    return range(*(arg for arg in (slc.start, slc.stop, slc.step) if arg is not None))


def get_mode_attr(attr: str, objs: list[Any]) -> tuple[bool, Any]:
    """
    Calculate the mode value of the attribute `attr`
    :param attr: attribute name
    :param objs: object list with attribute `attr`
    :return: success bool, then mode value
    """
    counts = []
    for obj in objs:
        if hasattr(obj, attr):
            this_value = getattr(obj, attr)

            for i, (value, count) in enumerate(counts):
                if value == this_value:
                    counts[i][1] += 1
                    break
            else:
                counts.append([this_value, 1])

    if counts:
        return True, max(counts, key=lambda x: x[1])[0]
    else:
        return False, None


def get_mode(objs: list[Any], no_dunder: bool = False):
    """
    Generate an object whose attributes are each a mode value of the inputted objects
    :param objs:
    :param no_dunder: whether to ignore __dunder__ attributes
    """
    if not objs:
        return None

    attrs = {}
    for attr in dir(objs[0]):
        if no_dunder and attr.startswith('__') and attr.endswith('__'):
            continue

        success, mode_attr = get_mode_attr(attr, objs)
        if success:
            try:
                mode_attr = copy.deepcopy(mode_attr)
            except Exception as e:
                warnings.warn(f"Could not deepcopy {mode_attr}.\nMaybe try setting `no_dunder` to True? Exception: {e}")
            attrs[attr] = mode_attr

    ret = object.__new__(type(objs[0]))
    ret.__dict__.update(attrs)
    return ret


def eval_inputs(soup: BeautifulSoup) -> dict[str, Any]:
    qs = {}

    # Parse inputs
    for _input in soup.find_all("input"):
        qs[_input.get("name")] = _input.get("value")

    # Parse select elements
    for select in soup.find_all("select"):
        options = select.find_all("option")

        # If there's no predefined value just pick the first
        selected_id = select.get("selected")
        selected = None
        for option in options:
            if not selected:
                if selected_id is not None:
                    if option.text == selected_id:
                        selected = option
                else:
                    selected = option

        qs[select.get("name")] = selected.get("value")

    return qs


def with_kwargs(cls):
    """
    Decorator for dataclasses to add a 'from_kwargs' method that ignores invalid kwargs

    for type hinting, add

    @classmethod
    def from_kwargs(cls, **kwargs) -> Self: ...
    """

    def from_kwargs(**kwargs):
        return cls(**{k: v for k, v in kwargs.items() if k in signature(cls).parameters})

    cls.from_kwargs = from_kwargs
    return cls



from __future__ import annotations

from typing import Any
from dataclasses import dataclass, field

from requests.structures import CaseInsensitiveDict

from . import session, digitalbook

@dataclass
class Course:
    id: int = None
    name: str = None
    subject: str = None
    smart: bool = field(repr=False, default=None)

    parent_id: int = field(repr=False, default=None)

    logo_url: str = field(repr=False, default=None)
    library_thumbnail_url: str = field(repr=False, default=None)
    course_name_image_url: str = field(repr=False, default=None)
    banner_background_image_url: str = field(repr=False, default=None)

    color: str = field(repr=False, default=None)
    banner_color: str = field(repr=False, default=None)
    lens_icon_color: str = field(repr=False, default=None)

    token_course: bool = field(repr=False, default=None)
    position: int = field(repr=False, default=None)
    self_study: Any = field(repr=False, default=None)
    no_contents_found_message: str = field(repr=False, default=None)
    curriculum: dict[str, str | dict[str, str | dict[str, str]]] = field(repr=False, default=None)

    _sess: session.Session = field(repr=False, default=None)

    def _get_img(self, url) -> tuple[bytes, CaseInsensitiveDict]:
        """
        Return img from url as a tuple: content, and headers
        """
        resp = self._sess.rq.get(f"https://www.kerboodle.com{url}")

        return resp.content, resp.headers

    @property
    def logo(self):
        return self._get_img(self.logo_url)

    @property
    def library_thumbnail(self):
        return self._get_img(self.library_thumbnail_url)
    @property
    def course_name_image(self):
        return self._get_img(self.course_name_image_url)
    @property
    def banner_background_image(self):
        return self._get_img(self.banner_background_image_url)

    @property
    def digital_books(self):
        data = self._sess.rq.get(f"https://www.kerboodle.com/api/courses/{self.id}/digital_books").json()

        return [digitalbook.DigitalBook.from_kwargs(_sess=self._sess, course=self, **attrs) for attrs in data]

    # --- * --- #



from __future__ import annotations

import atexit
import base64
import logging
import re
import time
import warnings
from dataclasses import dataclass
from typing import Optional

# noinspection PyProtectedMember
from playwright.sync_api import sync_playwright, PlaywrightContextManager, Playwright, Browser, Page, Request, \
    FilePayload
# noinspection PyProtectedMember
from playwright._impl import _errors as pw_errors

@dataclass
class Session:
    pw_ctx: PlaywrightContextManager
    playwright: Playwright
    browser: Browser
    page: Page

    def set_pfp(self, fp: str, *, return_screenshot: bool = True) -> Optional[bytes]:
        # depending on your theme, fetching the pfp seems to be inconsistent

        self.page.wait_for_load_state("networkidle", timeout=120_000)
        self.page.goto("https://myaccount.microsoft.com/")

        img_btn = self.page.locator("div[role='presentation'].ms-Persona-imageArea")
        img_btn.wait_for()

        # maybe change this to a locator: https://playwright.dev/python/docs/other-locators
        div = self.page.locator('div[role=heading]:has-text("Give feedback to Microsoft")')
        try:
            div.wait_for(timeout=10_000)
        except pw_errors.TimeoutError as e:
            warnings.warn(f"{e.__class__}: {e}")
            div = None

        if div:
            frame_parent = div.locator("../..")
            cancel_btn = frame_parent.locator('button:not([aria-disabled]):has-text("Cancel")')

            cancel_btn.click()

        img = img_btn.locator("img[src][alt='Profile photo']")
        try:
            img.wait_for()
        except pw_errors.TimeoutError as e:
            warnings.warn(f"Ignored: {e}")

        img_btn.click()
        change_photo_btn = self.page.wait_for_selector("span[role=presentation].CloudUpload")

        with self.page.expect_file_chooser() as fc_info:
            change_photo_btn.click()

        file_chooser = fc_info.value
        file_chooser.set_files(fp)

        self.page.wait_for_selector("button[aria-label=Save]").click()

        if return_screenshot:
            return self.page.screenshot()


def login(email: str, password: str, *, headless: bool=True, **kwargs):
    if not email.endswith("@kegs.org.uk"):
        email += "@kegs.org.uk"

    pw_ctx = sync_playwright()
    playwright = pw_ctx.__enter__()
    atexit.register(pw_ctx.__exit__)

    browser = playwright.chromium.launch(headless=headless, **kwargs)
    page = browser.new_page()

    page.goto("https://www.outlook.com/kegs.org.uk")

    email_input = page.wait_for_selector("input[type=email]")
    submit_input = page.wait_for_selector("input[type=submit]")

    email_input.type(email) # passwd
    submit_input.click()

    time.sleep(5)

    password_input = page.wait_for_selector("input[type=password]")
    submit_input = page.wait_for_selector("input[type=submit]")

    password_input.type(password)
    submit_input.click()

    page.wait_for_url("https://login.microsoftonline.com/common/login")
    page.wait_for_selector("input[type=submit]").click()

    # page.wait_for_url("https://login.microsoftonline.com/appverify")
    # try:
    #     page.wait_for_selector("input[type=submit]").click()
    # except Exception as e:
    #     warnings.warn(str(e))

    page.wait_for_load_state("networkidle", timeout=120_000)

    if page.url.startswith("https://login.microsoftonline.com/common/oauth2/v2.0/authorize"):
        table = page.wait_for_selector("div[class=table tole=button]")
        table.query_selector("div[class=table-row]").click()

    page.wait_for_url("https://outlook.office365.com/mail/")

    return Session(
        pw_ctx=pw_ctx, playwright=playwright, browser=browser, page=page
    )



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from datetime import datetime
from urllib.parse import urlparse, parse_qs
from typing import Any, Self

import dateparser
from bs4 import PageElement, BeautifulSoup

from . import session, user, tag
from ..util import commons, exceptions


@dataclass
class Event:
    title: str = None
    description: str = None
    location: str = None
    type: str = None

    date: datetime = None

    _sess: session.Session = None

@dataclass
class Calendar:
    """
    An instance of a calendar widget that KEGSNet can give
    """
    events: list[Event] = field(default_factory=list)

    _sess: session.Session = None



from __future__ import annotations

from dataclasses import dataclass, field
from urllib.parse import parse_qs, urlparse

from bs4 import BeautifulSoup, PageElement

from ..util import commons, exceptions

@dataclass
class Article:
    id: int
    title: str = None
    contents: PageElement = field(repr=False, default=None)

    @property
    def text(self):
        return self.contents.text.strip()

    def update_by_id(self):
        response = commons.REQ.get("https://it.kegs.org.uk/", params={
            "page_id": self.id
        })

        if not "page_id" in parse_qs(urlparse(response.url).query):
            raise exceptions.NotFound(f"'article' id={self.id} is not an article. (redirected to {response.url})")

        soup = BeautifulSoup(response.text, "html.parser")
        post = soup.find("div", {"id": "content"})

        try:
            heading = post.find("div", {"class": "singlepage"})

            self.title = heading.text
            self.contents = post
            # ^^ the reason why the whole post is used rather than only the 'entry' div is because of article #22,
            # which doesn't contain an entry div: https://it.kegs.org.uk/?page_id=22

        except AttributeError as e:
            raise exceptions.NotFound(f"article id={self.id} probably doesn't exist. error: {e}")


def get_article_by_id(_id: int):
    article = Article(_id)
    article.update_by_id()
    return article



import requests
from dataclasses import dataclass, field
from datetime import datetime
import dateparser
import mimetypes

from typing_extensions import deprecated

from urllib.parse import urlparse, parse_qs
from bs4 import BeautifulSoup, SoupStrainer

from ..util import commons, exceptions

@dataclass
class Asset:
    """
    Represents an asset that can be downloaded with the 'force_download.cfm' endpoint
    """

    id: int=None
    content: bytes = field(repr=False, default=None)
    name: str = None
    mime: str = None
    last_modified: datetime = None

    @property
    def ext(self):
        """
        Guess the file extension of this asset
        """
        return mimetypes.guess_extension(self.mime)

@deprecated("Downloading assets by ID has been patched in ~2025")
def download_asset_by_id(_id: int) -> Asset:
    """
    Fetch an asset by id from the force_download.cfm endpoint, using headers to provide metadata
    :param _id: id of asset to fetch
    :return: The corresponding Asset object
    """

    url = "https://www.kegs.org.uk/force_download.cfm"
    response = commons.REQ.get(url,
                            params={"id": _id})

    if response.url == f"{url}?id={_id}":
        raise exceptions.NotFound(f"Asset id {_id!r} could not be found (no redirect)")

    fname = urlparse(response.url).path.split('/')[-1]

    return Asset(_id,
                 response.content,
                 fname,
                 response.headers.get("Content-Type"),
                 dateparser.parse(response.headers.get("Last-Modified", '')))

@deprecated("Downloading assets by ID has been patched in ~2025. force_download.cfm is no longer used by the website.")
def find_asset_ids(url: str) -> list[int]:
    """
    Scrape any asset ids by looking for force_download.cfm links
    :param url: url of page to scrape
    :return: list of ids
    """
    ids = []

    global_netloc = urlparse(url).netloc

    links = commons.find_links(BeautifulSoup(
        commons.REQ.get(url).text, "html.parser", parse_only=SoupStrainer("a")
    ))

    for link in links:
        parse = urlparse(link)

        netloc = parse.netloc if parse.netloc else global_netloc

        if netloc == "www.kegs.org.uk" and parse.path == "/force_download.cfm":
            qparse = parse_qs(parse.query)
            ids.append(int(qparse.get("id")[0]))

    return ids



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from datetime import datetime

from bs4 import BeautifulSoup

from . import session, coursecategory
from ..util import exceptions


@dataclass
class Course:
    id: int = None
    display_name: str = None  # == fullname == fullnamedisplay
    name: str = None  # == shortname
    summary: BeautifulSoup = None  # summaryformat is always == 1 for me. maybe that affects this

    start_time: datetime = None
    end_time: datetime | None = None  # set to none if the value is 0
    access_time: datetime = None

    category: coursecategory.CourseCategory = None

    image: str = field(default=None, repr=False)

    hidden: bool = None

    _sess: session.Session = None

    """
    Should these attrs be ignored?:
    fullname
    viewurl
    progress
    hasprogress
    isfavourite
    showshortname
    """

    @classmethod
    def from_json(cls, data: dict[str, str | bool | int], sess: session.Session = None):
        if data["fullname"] != data["fullnamedisplay"]:
            warnings.warn(f"Please report to github, fullname != fullnamedisplay: {data}",
                          category=exceptions.UnimplementedWarning)

        return cls(
            _sess=sess,
            id=data["id"],
            display_name=data["fullnamedisplay"],
            name=data["shortname"],
            summary=BeautifulSoup(data["summary"], "html.parser"),
            start_time=datetime.fromtimestamp(data["startdate"]),
            end_time=datetime.fromtimestamp(data["enddate"]) if data.get("endate", 0) != 0 else None,
            access_time=datetime.fromtimestamp(data["timeaccess"]),
            image=data["courseimage"],
            hidden=data["hidden"],
            category=coursecategory.CourseCategory(_sess=sess,
                                                   name=data["coursecategory"])

        )



"""
Anything to do with the IT website (KEGS IT): https://it.kegs.org.uk/
"""
from .news import get_news_page, load_news_category, NewsItem, Category
from .images import download_header, download_banner
from .article import get_article_by_id, Article



from setuptools import setup
import os

requires = [req.strip() for req in open("requirements.txt").readlines()]

setup(
    name='kegscraper',
    version='v0.1.21',
    packages=['kegscraper'] +
             [f"kegscraper.{subdir}" for subdir in next(os.walk("kegscraper"))[1] if subdir != "__pycache__"],
    project_urls={
        "Homepage": 'https://kegs.org.uk/',
        "Source": "https://github.com/BigPotatoPizzaHey/kegscraper",
        "Documentation": "https://github.com/BigPotatoPizzaHey/kegscraper/wiki"
    },
    license=open("LICENSE").read(),
    author='BigPotatoPizzaHey',
    author_email="poo@gmail.com",
    description="The ultimate KEGS webscraping module",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=requires,
    data_files=[("/", ["requirements.txt"])]
)



from __future__ import annotations

import json
import warnings
from typing import Any, Self
from dataclasses import dataclass, field
from datetime import datetime
from urllib.parse import urlparse, urlunparse

import dateparser
from bs4 import BeautifulSoup
import lxml

from . import session, course

from ..util import commons


@dataclass
@commons.with_kwargs
class DigitalBook:
    id: int = None
    name: str = None

    published: bool = field(repr=False, default=None)
    is_new: bool = field(repr=False, default=None)
    is_updated: bool = field(repr=False, default=None)

    image_src: str = field(repr=False, default=None)
    content_object_link: str = field(repr=False, default=None)

    launcher: str = field(repr=False, default=None)
    # image_class: str

    purchase_url: dict[str, str] = field(repr=False, default=None)
    available: dict[str, str] = field(repr=False, default=None)
    purchased: dict[str, str] = field(repr=False, default=None)
    subs_end_date: datetime = field(repr=False, default=None)

    course: course.Course = field(repr=False, default=None)
    _sess: session.Session = field(repr=False, default=None)

    # engine: str
    # purchase_link: str

    # offline_content_link: Any
    # offline_content_version: int

    # url: dict[str, str]

    # purchase_link_text: dict[str, str]
    # purchase_instruction_text: dict[str, str]
    # purchase_popup_title: dict[str, str]

    def __post_init__(self):
        if not isinstance(self.subs_end_date, datetime) and self.subs_end_date is not None:
            self.subs_end_date = dateparser.parse(str(self.subs_end_date))

    @classmethod
    def from_kwargs(cls, **kwargs) -> Self:
        ...

    @property
    def url(self):
        return f"https://www.kerboodle.com/api/courses/{self.course.id}/interactives/{self.id}.html"

    @property
    def _interactive_html_data(self) -> dict | None:
        resp = self._sess.rq.get(self.url)
        soup = BeautifulSoup(resp.text, "html.parser")

        data = None
        to_find = '\n//<![CDATA[\n        window.authorAPI.setup('
        for script in soup.find_all("script"):
            if script.contents:
                js = script.contents[0]
                i = js.find(to_find)
                if i >= 0:
                    data = commons.consume_json(js, i + len(to_find))
                    break

        return data

    @property
    def _datajs_url(self) -> str:
        url = self._interactive_html_data["url"]
        parsed = urlparse(url)

        # remove index.html, add data.js instead

        path = '/'.join(parsed.path.split('/')[:-1] + ["data.js"])

        return str(urlunparse(parsed._replace(path=path)))

    @property
    def _datajs(self):
        assert lxml # you need lxml to parse xml

        resp = self._sess.rq.get(self._datajs_url)
        js = resp.text.strip()

        assert js.startswith("ajaxData = {")

        data: dict[str, str] = json.loads(js[len("ajaxData = "):-1])

        ret = {}
        for key, val in data.items():
            try:
                ret[key] = BeautifulSoup(val, "xml")
            except Exception as e:
                warnings.warn(f"Caught exception: {e}")
                ret[key] = val

        return ret

    @property
    def _catxml(self) -> BeautifulSoup:
        """
        :return: the other xml soup in datajs
        """
        xmldict = self._datajs

        return xmldict[next(filter(lambda x: x != "LearningObjectInfo.xml", xmldict))]

    @property
    def page_urls(self) -> list[str]:
        soup = self._catxml

        ret = []
        pages = soup.find("pages")
        for page in pages.find_all("page"):
            url = page.get("url")
            if url.startswith("//"):
                url = f"https:{url}"

            ret.append(url)

        return ret



"""
Get news page & load news category functions & related dataclasses
"""

from __future__ import annotations

from urllib.parse import parse_qs, urlparse

from dataclasses import dataclass, field
from datetime import datetime

from bs4 import BeautifulSoup, Comment

from ..util import commons, exceptions


@dataclass
class Category:
    """
    A dataclass representing a category of a news item on kegsIT
    """
    id: int
    name: str


@dataclass
class NewsItem:
    """
    A news 'article'/post on the kegsIT website
    """
    id: int

    author: str = None
    title: str = None
    content: str = field(repr=False, default=None)

    date: datetime = None
    category: Category = None


def get_news_page(page: int = 1, category: int | Category = 7) -> NewsItem:
    """
    Get the news item at the specified page index
    :param category: Category of news
    :param page: Page index
    :return: The news item
    """
    if isinstance(category, Category):
        category = category.id

    # Find the page corresponding to the category & post
    response = commons.REQ.get(f"https://it.kegs.org.uk/",
                               params={
                                   "cat": category,
                                   "paged": page
                               })

    if response.status_code == 404:
        raise exceptions.NotFound(f"Could not find news page. Content: {response.content}")

    text = response.text
    soup = BeautifulSoup(text, "html.parser")

    anchor = soup.find("a", {"rel": "bookmark"})

    url = anchor.attrs.get("href")
    qparse = parse_qs(urlparse(url).query)

    news_id = int(qparse["p"][0])

    # Actually scrape the main page for this news item
    text = commons.REQ.get("https://it.kegs.org.uk/",
                           params={
                               "p": news_id
                           }).text
    soup = BeautifulSoup(text, "html.parser")

    title = soup.find("div", {"class": "singlepage"}).text

    date_elem = soup.find("abbr")
    date = datetime.fromisoformat(date_elem.attrs.get("title"))

    # There is a cheeky comment above the date element that we can try to webscrape
    comment = date_elem.parent.find(string=lambda _text: isinstance(_text, Comment)).extract()
    # It's actually in HTML format
    comment_text = BeautifulSoup(comment, "html.parser").text

    author = commons.webscrape_value(comment_text, "Written by ", " on")

    # Contents and category
    contents_div = soup.find("div", {"id": "content"})

    post_wrapper = contents_div.find("div", {"id": "singlepostwrapper"})
    category_anchor = post_wrapper.find("a", {"rel": "category"})

    # url = category_anchor.attrs.get("href")  # We already know the category index so we don't need to parse the link
    category_name = category_anchor.text
    category_obj = Category(category, category_name)

    # Get content
    content = contents_div.find("div", {"class": "entry"}).text.strip()

    return NewsItem(
        news_id,
        author,

        title,
        content,

        date,
        category_obj
    )


def load_news_category(category: int | Category = 7, *, limit: int = 10, offset: int = 0) -> list[NewsItem]:
    """
    Make mutliple requests to kegsIT to load an entire category of news data with a given offset and limit
    :param category: Category of news data to scrape. Defaults to the 'news' category
    :param limit: # of Posts to scrape
    :param offset: Starting post index
    :return: A list of posts aka news items
    """
    pages = []
    for page in commons.generate_page_range(limit, offset, 1, 1)[0]:
        # print(page)
        try:
            pages.append(get_news_page(page, category))

        except exceptions.NotFound:
            break

    return pages



from __future__ import annotations

from dataclasses import dataclass, field, InitVar

from . import session

@dataclass
class Subject:
    name: str = None
    short_name: str = field(repr=False, default=None)

    image_url: str = None

    _session: session.Session = field(repr=False, default=None)

@dataclass
class Series:
    name: str = None
    id: int = None
    subject: Subject = None

    _session: session.Session = field(repr=False, default=None)

    subject_name: InitVar[str] = None
    subject_short_name: InitVar[str] = None

    def __post_init__(self, subject_name: str, subject_short_name: str):
        if not self.subject:
            self.subject = Subject()

        self.subject.name = subject_name
        self.subject.short_name = subject_short_name

        self.subject._session = self._session

    def get_books(self):
        self._session.page.goto(f"https://www.pearsonactivelearn.com/app/library/series/view/{self.id}#/studentbooks")

        self._session.page.locator(selector:="ul .tab-book-list.clearfix.studentbooks]").wait_for()
        print(self._session.page.query_selector(selector))


import requests
from bs4 import BeautifulSoup

from ..util import commons

def api_fetch(url: str, rq: requests.Session = None):
    """Make an api call to the oliver library api."""

    if rq is None:
        rq = commons.REQ

    url = f"https://kegs.oliverasp.co.uk/library/home/api/{url}"
    # For some reason it seems that it provides us with some query params then redirects us
    text = commons.REQ.get(url).text
    soup = BeautifulSoup(text, "html.parser")

    qs = {}

    # Parse inputs
    inputs = soup.find_all("input")
    for _input in inputs:
        attrs = _input.attrs
        qs[attrs.get("name")] = attrs.get("value")

    # Parse select elements
    selects = soup.find_all("select")
    for select in selects:
        attrs = select.attrs
        options = select.find_all("option")

        # If there's no predefined value just pick the first
        selected_id = attrs.get("selected")
        selected = None
        for option in options:
            if not selected:
                if selected_id is not None:
                    if option.text == selected_id:
                        selected = option
                else:
                    selected = option

        qs[attrs.get("name")] = selected.attrs.get("value")

    return rq.get(url, params=qs).json()



from __future__ import annotations

import atexit
import math
import random
import string
import enum
import time
import warnings
from typing import Any
from dataclasses import dataclass, field
from urllib.parse import quote_plus, ParseResult, urlunparse, urlencode

from . import series
from ..util import commons, exceptions

# noinspection PyProtectedMember
from playwright.sync_api import sync_playwright, PlaywrightContextManager, Playwright, Browser, Page, Request


class EventType(enum.Enum):
    library_update = "library_update"


@dataclass
class DebugSettings:
    print_on_req: bool = False


@dataclass
class Session:
    pw_ctx: PlaywrightContextManager = field(repr=False)
    playwright: Playwright = field(repr=False)
    browser: Browser = field(repr=False)
    page: Page = field(repr=False)

    events: dict[str, Any] = field(
        default_factory=dict)  # special dict that logs if certain events occur - e.g. updating based on library data
    debug: DebugSettings = field(repr=False, default_factory=DebugSettings)

    def __post_init__(self):
        self.page.on("requestfinished", self.on_req)

        atexit.register(self.pw_ctx.__exit__)

    def _register_event(self, event_type: EventType, value: Any = True):
        self.events[event_type.value] = value

    def _expect_event(self, event_type: EventType, *, pop: bool = True, timeout: float | int = math.inf):
        start = time.time()
        while event_type.value not in self.events:
            if time.time() > start + timeout:
                raise exceptions.TimeOut(f"Ran out of time waiting for {event_type.value!r} after {timeout}s")

        if pop:
            return self.events.pop(event_type.value)
        else:
            return self.events[event_type.value]

    def clear_events(self):
        self.events.clear()

    def on_req(self, req: Request):
        _ = req.url, str(req)  # it appears that calling this property & __str__ affects behaviour, which is very odd

        resp = req.response()

        if self.debug.print_on_req:
            print(f"Session.on_req.{req = }")

        match req.url:
            case "https://www.pearsonactivelearn.com/app/Execute/GetUserSeriesList":
                data: dict[str, Any] = resp.json()

                # 4 items in each list?
                data: list[dict] = data["Data"]
                for obj in data:
                    columns = obj["Columns"]
                    if columns == ['SeriesName', 'SeriesId', 'Subject', 'SubjectId']:
                        data: list[list[str | int]] = obj["Data"]
                        break
                else:
                    warnings.warn(f"Failed to parse {data}", category=exceptions.ParseFailure)

                self._register_event(EventType.library_update, [
                    series.Series(name=series_data[0],
                                  id=series_data[1],
                                  subject_name=series_data[2],
                                  subject_short_name=series_data[3],
                                  _session=self) for series_data in data
                ])

    @property
    def library(self) -> list[series.Series]:
        self.clear_events()
        self.page.goto("https://www.pearsonactivelearn.com/app/library")
        return self._expect_event(EventType.library_update)


def login(username: str, password: str, headless: bool = True, **kwargs):
    # we can generate a random ies code by ourselves. We can go directly to the pearson login page
    iescode = ''.join(random.choices(string.ascii_letters + string.digits, k=10))
    url = urlunparse(ParseResult(scheme='https',
                                 netloc='login.pearson.com',
                                 path='/v1/piapi/iesui/signin',
                                 params='',
                                 query=urlencode({
                                     # client id is always the same
                                     "client_id": "BRSIcPHr2Iq0NV8AQP99zDZau8IPUxgy",
                                     "redirect_uri": "https://www.pearsonactivelearn.com/app/login",
                                     "nonce": "123454321",
                                     "prompt": "login",
                                     "cookieConsentGroups": "100000",
                                     "login_success_url": f"https://www.pearsonactivelearn.com/app/login?iesCode={quote_plus(iescode)}"
                                 }),
                                 fragment=''))  # type:ignore

    pw_ctx = sync_playwright()
    playwright = pw_ctx.__enter__()

    browser = playwright.webkit.launch(headless=headless, **kwargs)
    page = browser.new_page()

    page.goto(str(url))

    username_inp = page.wait_for_selector("input[id=username]")
    password_inp = page.wait_for_selector("input[id=password]")
    submit_btn = page.wait_for_selector("button[id=submitBttn]")

    username_inp.type(username)
    password_inp.type(password)
    submit_btn.click()

    page.wait_for_url("https://www.pearsonactivelearn.com/app/library")

    return Session(
        pw_ctx=pw_ctx,
        playwright=playwright,
        browser=browser,
        page=page
    )
